# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import json

from agml.framework import AgMLSerializable
from agml.backend.config import data_save_path
from agml.utils.downloads import download_dataset
from agml.utils.io import get_file_list, get_dir_list, is_image_file


class DataBuilder(AgMLSerializable):
    """Builds an internal representation format of AgML data.

    This class doesn't affect the state of the `AgMLDataLoader`, but
    instead generates an internal representation of the data content
    of a dataset. This allows for all data to be loaded in a standard
    format and streamlined into the `DataManager`.

    Primarily, this class attempts to locate the dataset and if unable
    to do so, it downloads the dataset from the public bucket. Then,
    it creates a mapping between the images and annotations which is
    used by the `DataManager` inside the `AgMLDataLoader`.
    """
    serializable = frozenset(
        ('name', 'labels_for_image', 'data',
         'info', 'dataset_root', 'data_length'))

    def __init__(self, info, dataset_path, overwrite):
        # Attempt to locate or download the dataset.
        self._info = info
        self._name = info.name
        self._configure_dataset(
            dataset_path = dataset_path, overwrite = overwrite)
        self._data = None
        self._data_length = None
        self._labels_for_image = None

    @classmethod
    def from_data(cls, contents, info, root, builder):
        """Initializes the `DataBuilder` from a pre-built set of data.

        This is mainly used when running `AgMLDataLoader.split`, to create
        a `DataManager` with a split of data. Functionally, it disables
        all of the actual generation protocols of the `DataBuilder` and
        just pre-assigns all of the values to it.
        """
        obj = super(DataBuilder, cls).__new__(cls)
        obj._name = info.name
        obj._info = info
        obj._dataset_root = root
        obj._data = contents[0]
        obj._data_length = len(obj._data)
        obj._labels_for_image = contents[1]
        if hasattr(builder, '_default_coco_annotations'):
            obj._default_coco_annotations = DataBuilder._regenerate_coco_annotations(
                builder._default_coco_annotations, obj._data)
        return obj

    @staticmethod
    def _regenerate_coco_annotations(coco, data):
        """Regenerates COCO annotations from a reduced set."""
        image_coco = []
        annotation_coco = []
        image_id_tracker = 0

        # Update the list of image paths.
        for c in coco['images']:
            if c['file_name'] in list(data.keys()):
                c['image_id'] = image_id_tracker
                image_coco.append(c)
                image_id_tracker += 1

        # Update the image IDs for the annotations
        for image_id, (_, annotation) in enumerate(data.items()):
            for a in annotation:
                a['id'] = image_id
            annotation_coco.append(annotation)

        # Update the complete COCO dictionary.
        ret_coco = coco.copy()
        ret_coco['images'] = image_coco
        ret_coco['annotations'] = annotation_coco
        return ret_coco

    @property
    def dataset_root(self):
        return self._dataset_root

    def _configure_dataset(self, **kwargs):
        """Finds and configures the dataset into the loader."""
        # Check if the user wants to overwrite the existing dataset,
        # and resolve the potentially provided custom dataset path.
        overwrite = kwargs.get('overwrite', False)
        if kwargs.get('dataset_path', None):
            kwargs['dataset_path'] = os.path.realpath(
                os.path.expanduser(kwargs['dataset_path']))

        # If a custom dataset has been provided, then we assume the directory
        # provided is the root of the dataset, e.g., if /root/datasets/dir
        # is provided, then dir doesn't necessarily need to be the name of
        # the dataset, but it is assumed to be its root. Note that if no path
        # is given, then we check similarly to the default path for datasets,
        # e.g., if there is a directory with the provided name in the custom path.
        #
        # No downloading phase for custom datasets, as there is nothing to do.
        if self._info.is_custom_dataset:
            if kwargs.get('dataset_path', False):
                path = kwargs.get('dataset_path')
                if os.path.exists(path):
                    self._dataset_root = path
                    return

            elif os.path.exists(os.path.join(
                    data_save_path(), self._name)):
                self._dataset_root = os.path.join(
                    data_save_path(), self._name)
                return

            # Otherwise, we don't know what to do.
            raise OSError(
                f"Could not find a directory for the dataset {self._name}. "
                f"Dataset path given: {kwargs.get('dataset_path', None)}")

        # If the user doesn't want to overwrite the existing contents,
        # first check whether the dataset already exists. If so, then set
        # the dataset root and return without doing any downloading.
        #
        # Note that the `dataset_path` is resolved as follows: If the path
        # ends with the name of the dataset, e.g., `/root/datasets/<name>`,
        # then we pop '<name>' from the end of the path in order to prevent
        # the dataset from being downloaded to `/root/datasets/<name>/<name>`.
        if not overwrite:
            if kwargs.get('dataset_path', False):
                path = kwargs.get('dataset_path')
                if (os.path.basename(path) == self._name
                    and os.path.exists(path)) or \
                        os.path.exists(os.path.join(path, self._name)):
                    if os.path.basename(path) != self._name:
                        path = os.path.join(path, self._name)
                    self._dataset_root = path
                    return

            elif os.path.exists(os.path.join(
                    data_save_path(), self._name)):
                self._dataset_root = os.path.join(
                    data_save_path(), self._name)
                return

        # If the user wants to overwrite, or the dataset doesn't exist
        # at the path, then we create the root in the same way, except
        # we just also download the dataset to the path.
        else:
            if kwargs.get('dataset_path', False):
                path = kwargs.get('dataset_path')
                if (os.path.basename(path) == self._name
                    and os.path.exists(path)) or \
                        os.path.exists(os.path.join(path, self._name)):
                    if os.path.basename(path) != self._name:
                        path = os.path.join(path, self._name)
                    print(f'[AgML Download]: Overwriting dataset at '
                          f'{os.path.join(path)}')
                    if os.path.basename(path) != self._name:
                        path = os.path.join(path, self._name)
                    self._dataset_root = path

            elif os.path.exists(os.path.join(
                    data_save_path(), self._name)):
                self._dataset_root = os.path.join(
                    data_save_path(), self._name)
                sys.stderr.write(
                    f'[AgML Download]: Overwriting dataset at '
                    f'{os.path.join(self._dataset_root)}')

        # Performs the actual downloading of the dataset.
        if kwargs.get('dataset_path', False):
            download_path = kwargs['dataset_path']
            if os.path.basename(download_path) != self._name:
                download_path = os.path.join(download_path, self._name)
        else:
            download_path = os.path.join(data_save_path(), self._name)
        sys.stderr.write(
            f"[AgML Download]: Downloading dataset "
            f"`{self._name}` to {download_path}.")
        download_dataset(self._name, download_path)
        self._dataset_root = download_path

    def _generate_contents(self, task):
        """Dispatches to a content generation method for the provided task."""
        if self._data is not None:
            return
        if task == 'image_classification':
            self._generate_image_classification_data()
        elif task == 'image_regression':
            self._generate_image_regression_data()
        elif task == 'object_detection':
            self._generate_object_detection_data()
        else:
            self._generate_semantic_segmentation_data()

    def get_contents(self):
        """Extracts the internal representation of the data content."""
        # Create the internal content representation of the dataset.
        self._generate_contents(self._info.tasks.ml)

        # Update the length of the data in the metadata dictionary for
        # custom datasets, since they may not be providing the total number.
        self._data_length = len(self._data)
        if self._info._metadata.get('n_images', None) is None:
            self._info._metadata['n_images'] = str(self._data_length)
        return self._data

    def export_contents(self, export_format):
        """Returns the raw contents of the loader."""
        # We start by constructing a default mapping.
        contents = self.get_contents()

        # For a COCO JSON dictionary, we have to make the full paths.
        if self._info.tasks.ml == 'object_detection':
            paths, coco = contents.keys(), contents.values()
            paths = [os.path.join(
                self._dataset_root, 'images', i) for i in paths]
            contents = dict({k: v for k, v in zip(paths, coco)})

        # If the export format is `None`, we return the default mapping.
        if export_format is None:
            return contents

        # If the export format is `arrays`, return the keys and
        # the values of the mapping as two independent arrays.
        if export_format == 'arrays':
            return list(contents.keys()), list(contents.values())

        # A special case for COCO JSON dictionaries.
        if export_format == 'coco':
            if self._info.tasks.ml != 'object_detection':
                raise ValueError("The `coco` export format is "
                                 "only for object detection tasks.")
            return self._default_coco_annotations

    # The following methods actually generate the content mappings for
    # the different tasks. In essence, for each image path, `image`, a
    # mapping is generated with a corresponding annotation, such as a
    # label for image classification, mask for semantic segmentation, or
    # a COCO JSON dictionary for object detection.

    def _generate_image_classification_data(self):
        """Loads image classification data for the `directory_names` format.

        In this format, images are organized by class where the directory
        they are placed in corresponds to their label in the dataset.
        """
        image_label_mapping = {}
        candidate_dirs = get_dir_list(self._dataset_root)
        for dir_ in candidate_dirs:
            if dir_.startswith('.'):
                continue
            dir_path = os.path.join(self._dataset_root, dir_)
            if len(get_file_list(dir_path)) == 0:
                continue
            for file_ in get_file_list(dir_path):
                file_ = os.path.join(dir_path, file_)
                image_label_mapping[file_] = self._info.class_to_num[dir_]
        self._data = image_label_mapping

    def _generate_image_regression_data(self):
        """Loads image regression data for the loader.

        In this format, there are input images in an `images` folder as
        well as other image formats in other various `*_images` folders,
        and an `annotations.json` file containing the regression outputs.
        """
        with open(os.path.join(self._dataset_root, 'annotations.json'), 'r') as f:
            annotations = json.load(f)
        content_mapping = {'inputs': [], 'outputs': []}
        annotation_types = set(list(self._info.class_to_num.keys()))
        annotation_types.remove('regression')
        for sample in annotations:
            for k, v in sample.items():
                if is_image_file(v):
                    sample[k] = os.path.join(self._dataset_root, f'{k}s', v)
            content_mapping['inputs'].append({
                k: v for k, v in sample.items()
                if re.match('(.*?)image', k)})
            out = {'regression': list(sample['outputs']['regression'].values())}
            out.update({
                k: self._info.class_to_num[k][v] for k, v in sample['outputs'].items()
                if k in annotation_types})
            content_mapping['outputs'].append(out)
        self._data = content_mapping

    def _generate_semantic_segmentation_data(self):
        """Loads semantic segmentation data for the loader.

        Image data is loaded from an `images` directory, and pixel-wise
        annotated images are loaded from an `annotations` directory.
        """
        image_dir = os.path.join(self._dataset_root, 'images')
        annotation_dir = os.path.join(self._dataset_root, 'annotations')
        images, annotations = sorted(get_file_list(image_dir)), \
                              sorted(get_file_list(annotation_dir))
        image_annotation_map = {}
        for image_path, annotation_path in zip(images, annotations):
            image_annotation_map[os.path.join(image_dir, image_path)] \
                = os.path.join(annotation_dir, annotation_path)
        self._data = image_annotation_map

    def _generate_object_detection_data(self):
        """Generates object detection data for the loader.

        Image data is loaded from an `images` directory, and the COCO
        JSON annotations are loaded from an `annotations.json` file.
        """
        with open(os.path.join(self._dataset_root, 'annotations.json')) as f:
            self._default_coco_annotations = json.load(f)
        coco_annotations = self._default_coco_annotations
        image_id_mapping = {}
        for img_meta in coco_annotations['images']:
            image_id_mapping[img_meta['id']] = img_meta['file_name']
        coco_map = {fname: [] for fname in image_id_mapping.values()}
        image_category_map = {k: [] for k in coco_map.keys()}
        for a_meta in coco_annotations['annotations']:
            image_category_map[image_id_mapping[a_meta['image_id']]] \
                .append(a_meta['category_id'])
            coco_map[image_id_mapping[a_meta['image_id']]].append(a_meta)
        self._labels_for_image = image_category_map
        self._data = coco_map




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from agml.backend.tftorch import tf
from agml.data.managers.training import TrainState
from agml.data.object import DataObject
from agml.utils.logging import log


class TFExporter(object):
    """Exports an `AgMLDataLoader` as a `tf.data.Dataset`.

    This class manages the methods and conversions between `AgMLDataLoader`
    parameters, managers, and methods, into TensorFlow native methods.
    """
    def __init__(self, task, builder):
        self._task = task
        self._builder = builder

    def digest_transforms(self, transforms, resizing):
        """Parses the transforms for the `AgMLDataLoader`."""
        self._transforms = {
            k: state for k, state in transforms.items()}
        self._size = resizing if resizing is not None else (512, 512)

    def assign_state(self, state):
        """Updates the training state for the `tf.data.Dataset`."""
        if state in [TrainState.EVAL,
                     TrainState.EVAL_TF,
                     TrainState.EVAL_TORCH,
                     TrainState.FALSE]:
            self._state = 'eval'
        else:
            self._state = 'train'

    def _build_from_contents_by_type(self, builder):
        """Parses the provided mapping into a valid set of contents."""
        if self._task == 'image_classification':
            ds = self._build_image_classification(builder)
            return ds.map(self._image_classification_load)
        elif self._task == 'semantic_segmentation':
            ds = self._build_semantic_segmentation(builder)
            return ds.map(self._semantic_segmentation_load)
        else:
            ds = self._build_object_detection(builder)
            return ds.map(self._object_detection_load)

    def _apply_resizing_by_type(self, ds):
        """Applies resizing based on the task."""
        if self._task == 'image_classification':
            return ds.map(self._image_classification_resize)
        elif self._task == 'semantic_segmentation':
            return ds.map(self._semantic_segmentation_resize)
        else:
            return ds.map(self._object_detection_resize)

    def build(self, batch_size = None):
        """Builds the `tf.data.Dataset` using the provided parameters."""
        # Construct the dataset from the contents.
        ds = self._build_from_contents_by_type(self._builder)

        # Apply the digested transforms and resizing.
        ds = self._apply_resizing_by_type(ds)
        if self._state != 'eval':
            # No transforms for object detection, since it is near impossible
            # for TensorFlow's graph mode to use COCO JSON dictionaries.
            if self._task == 'object_detection':
                if len(self._transforms) != 0:
                    log("Got transforms when exporting an `AgMLDataLoader`"
                        "to a `tf.data.Dataset`. These transforms will not be "
                        "applied. To use transforms in TensorFlow, use the "
                        "`as_keras_sequence()` method instead.")
            else:
                tfm = self._apply_transforms
                if len(self._transforms) != 0:
                    ds = ds.map(tfm)

        # Apply batching and prefetching, then return the dataset.
        if batch_size is not None:
            ds = ds.batch(batch_size)
            ds = ds.prefetch(batch_size)
        return ds

    # The following methods are used to parse the input
    # contents into valid methods for the loaders.

    @staticmethod
    def _build_image_classification(builder):
        images, labels = builder.export_contents(
            export_format = 'arrays')
        images, labels = tf.constant(images), tf.constant(labels)
        ds = tf.data.Dataset.from_tensor_slices((images, labels))
        return ds.shuffle(len(images))

    @staticmethod
    def _build_semantic_segmentation(builder):
        images, masks = builder.export_contents(
            export_format = 'arrays')
        images, masks = tf.constant(images), tf.constant(masks)
        ds = tf.data.Dataset.from_tensor_slices((images, masks))
        return ds.shuffle(len(images))

    @staticmethod
    def _build_object_detection(builder):
        images, annotations = builder.export_contents(
            export_format = 'arrays')
        images = tf.constant(images)
        processed_annotations = [
            DataObject._parse_coco(a) for a in annotations]
        features = {'bbox': [], 'category_id': [], 'area': [],
                    'image_id': [], 'iscrowd': [], 'segmentation': []}
        for a_set in processed_annotations:
            for feature in features.keys():
                features[feature].append(a_set[feature]) # noqa
        for feature in features.keys():
            features[feature] = tf.ragged.constant(features[feature])
        feature_ds = tf.data.Dataset.from_tensor_slices(features)
        ds = tf.data.Dataset.zip((
            tf.data.Dataset.from_tensor_slices(images), feature_ds))
        return ds.shuffle(len(images))

    # The following methods are used for loading images and
    # annotations for each of the different types of tasks.

    @staticmethod
    @tf.function
    def _image_classification_load(image, label):
        image = tf.cast(tf.image.decode_jpeg(
            tf.io.read_file(image)), tf.float32) / 255.
        return image, tf.convert_to_tensor(label)

    @staticmethod
    @tf.function
    def _semantic_segmentation_load(image, mask):
        image = tf.cast(tf.image.decode_jpeg(
            tf.io.read_file(image)), tf.float32) / 255.
        mask = tf.image.decode_jpeg(tf.io.read_file(mask))
        return image, mask

    @staticmethod
    @tf.function
    def _object_detection_load(image, coco):
        image = tf.cast(tf.image.decode_jpeg(
            tf.io.read_file(image)), tf.float32) / 255.
        ret_coco = coco.copy()
        for key in coco.keys():
            try: ret_coco[key] = coco[key].to_tensor()
            except: pass
        return image, ret_coco

    # The following methods apply resizing to the data.

    def _image_classification_resize(self, image, label):
        image = (tf.image.resize(
            image, self._size, method = 'nearest'), tf.float32)
        return image, label

    def _semantic_segmentation_resize(self, image, mask):
        image = tf.cast(tf.image.resize(
            image, self._size, method = 'nearest'), tf.float32)
        mask = tf.cast(tf.image.resize(
            mask, self._size, method = 'nearest'), tf.float32)
        return image, mask

    def _object_detection_resize(self, image, coco):
        # Helper for the `tf.py_function` for object detection.
        def _resize_image_and_bboxes(image, coco_boxes):
            nonlocal size
            y_scale, x_scale = image.shape[0:2]
            stack_boxes = tf.stack(
                [coco_boxes[:, 0] / x_scale,
                 coco_boxes[:, 1] / y_scale,
                 coco_boxes[:, 2] / x_scale,
                 coco_boxes[:, 3] / y_scale], axis = -1)
            image = tf.cast(tf.image.resize(
                image, size), tf.float32)
            y_new, x_new = image.shape[0:2]
            new_stack = tf.cast(tf.stack(
                [stack_boxes[:, 0] * x_new,
                 stack_boxes[:, 1] * y_new,
                 stack_boxes[:, 2] * x_new,
                 stack_boxes[:, 3] * y_new], axis = -1
            ), tf.int32)
            areas = new_stack[:, 2] * new_stack[:, 3]
            return image, new_stack, areas

        # The actual resizing can't take place in graph mode, so we
        # dispatch to a `tf.py_function` to do the resizing, then
        # re-assign the values back to the COCO JSON dictionary.
        size = self._size
        image, ret_coco_boxes, ret_areas = tf.py_function(
            _resize_image_and_bboxes,
            [image, coco['bbox']],
            [tf.float32, tf.int32, tf.int32])
        coco['bbox'] = ret_coco_boxes
        coco['area'] = ret_areas
        return image, coco

    # The following method manages the application of transforms.

    def _apply_transforms(self, image, annotation):
        # Helper for the `tf.py_function` for most transforms.
        def _py_apply(img, ann):
            nonlocal transforms
            img, ann = img.numpy(), ann.numpy()
            for key, state in transforms.items():
                if key == 'transform':
                    for t in state:
                        img = t(img)
                elif key == 'target_transform':
                    for t in state:
                        ann = t(ann)
                else:
                    for t in state:
                        img, ann = t(img, ann)

            return img, ann

        # The actual transforming can't take place in graph mode
        # (in most cases), so we dispatch and reassign.
        transforms = self._transforms
        image, annotation = tf.py_function(
            _py_apply,
            [image, annotation],
            [tf.float32, tf.int32])
        return image, annotation












# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import numpy as np

from agml.framework import AgMLSerializable
from agml.utils.random import seed_context


class TransformApplierBase(AgMLSerializable):
    """Applies a transform to the input data.

    This is used as a wrapper class for applying transformations in
    more complex cases. Derived classes implement the `apply` method
    to wrap the input transformation.

    This is used in cases such as `albumentations` transforms, which
    require keyword arguments. This wrapper allows them to be used
    as a traditional method for consistency in application.
    """
    serializable = frozenset(('transform', ))

    def __init__(self, transform):
        self._transform = transform

    @abc.abstractmethod
    def apply(self, *args, **kwargs):
        """Applies the transformation to the input data."""
        return

    def __call__(self, *args):
        return self.apply(*args)

    def __str__(self):
        return self.__class__.__name__ + f": {self._transform}"


class AlbumentationsTransformSingle(TransformApplierBase):
    def apply(self, image):
        transform = self._transform(image = image)
        return transform['image']


class AlbumentationsTransformMask(TransformApplierBase):
    def apply(self, image, mask):
        transform = self._transform(image = image, mask = mask)
        return transform['image'], transform['mask']


class AlbumentationsTransformCOCO(TransformApplierBase):
    def apply(self, image, coco):
        # This method is a bit more complex. We can't just apply it
        # to the COCO JSON dictionary as we need to extract the
        # bounding boxes/category IDs, do the transformation on those,
        # and then re-insert them into the COCO dictionary.
        coco_boxes = coco['bbox']
        coco_labels = coco['category_id']
        bboxes = np.c_[coco_boxes, coco_labels]
        transform = self._transform(image = image, bboxes = bboxes)
        image, bboxes = transform['image'], \
                        np.array(transform['bboxes'])[:, :-1]
        coco['bbox'], areas = bboxes, []
        for box in bboxes:
            areas.append(
                (box[0] + box[2]) * (box[1] + box[3]))
        coco['area'] = areas.copy()
        return image, coco


class SameStateImageMaskTransform(TransformApplierBase):
    def apply(self, image, mask):
        # This method applies a transformation to the image and mask.
        # Essentially, we need to set the random seed before applying
        # it to the image and mask to get the same results for both
        # the image and the mask. So, we change the seed each time, but
        # we also reset the seed so that it goes back to normal afterwards.
        seed = np.random.randint((1 << 31) - 1) # default is 32-bit systems
        with seed_context(seed) as context:
            image = self._transform(image)
            context.reset()
            mask = self._transform(mask)
        return image, mask


class NormalizationTransformBase(TransformApplierBase, abc.ABC):
    """A subclass to mark transforms as normalizing transforms."""
    pass


class ScaleTransform(NormalizationTransformBase):
    def apply(self, image):
        if image.max() >= 1 or np.issubdtype(image.dtype, np.integer):
            image = (image / 255).astype(np.float32)
        return image


class NormalizationTransform(NormalizationTransformBase):
    def apply(self, image):
        # This method applies normalization to input images, scaling them
        # to a 0-1 float range, and performing scaling normalization.
        if image.max() >= 1 or np.issubdtype(image.dtype, np.integer):
            image = (image / 255).astype(np.float32)

        mean, std = self._transform
        mean = np.array(mean, dtype = np.float32)
        std = np.array(std, dtype = np.float32)
        denominator = np.reciprocal(std, dtype = np.float32)
        image = (image - mean) * denominator
        return image


class OneHotLabelTransform(TransformApplierBase):
    def apply(self, labels):
        # This applies a one-hot label transformation. The only argument
        # in the `_transform` parameter is simply the number of labels.
        one_hot = np.zeros(shape = (self._transform, ))
        one_hot[labels] = 1
        return one_hot.astype(np.float32)


class MaskToChannelBasisTransform(TransformApplierBase):
    def apply(self, mask):
        # Converts a 2-d mask of `n` labels, excluding background, to a
        # `n x h x w` 3-dimensional mask with each channel a label.
        input_shape = mask.shape
        mask = mask.ravel()
        n = mask.shape[0]
        mask = np.array(mask, dtype = np.int32)
        out = np.zeros(shape = (n, self._transform + 1))
        out[np.arange(n), mask] = 1
        out = np.reshape(out, input_shape + (self._transform + 1,))
        return out[..., 1:].astype(np.int32)





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from agml.backend.tftorch import torch, tf, get_backend


def set_seed(seed = None):
    """Sets a new random seed. If None, uses a random seed."""
    import random
    random.seed(seed)
    import numpy as np
    np.random.seed(seed)
    if get_backend() == 'torch':
        torch.random.manual_seed(seed)
    elif get_backend() == 'tensorflow':
        tf.random.set_seed(seed)




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import glob
import json
import shutil
from typing import List
from dataclasses import dataclass
from datetime import datetime as dt

import cv2
import numpy as np

from agml.utils.io import recursive_dirname, get_dir_list, get_file_list
from agml.utils.logging import tqdm


@dataclass
class DataFormatConverterMetadata:
    path: str
    name: str
    image_size: str
    annotation_type: str
    labels: List[str]
    generation_date: str


class HeliosDataFormatConverter(object):
    """Converts the annotation and organization of a Helios dataset.

    This method can be used to convert a Helios-generated dataset into the
    corresponding AgML format for the annotation type, thus allowing the dataset
    to be used with an `AgMLDataLoader`. The conversion consists of replacing the
    file structure in the directory location of the dataset with the new structure,
    and adding a new file at `.metadata/agml_info.json` which contains as much info
    about the dataset that can be gleaned in order to have loader compatibility.

    Parameters
    ----------
    dataset : str
        The name of the dataset to convert.

    Notes
    -----
    This class shouldn't be instantiated on its own, instead use the helper method
    `agml.synthetic.convert_helios_dataset_format()` with the dataset.
    """

    def __init__(self, dataset):
        # Locate the dataset and parse its metadata.
        self._meta = self.parse_metadata(dataset)

    def parse_metadata(self, dataset):
        """Parse the metadata of the dataset for use in conversion."""
        # Locate the dataset and get its actual name.
        output_dir = self._locate_dataset(dataset)
        name = os.path.basename(output_dir)
        meta = {'path': output_dir, 'name': name}
        meta_dir = os.path.join(output_dir, '.metadata')

        # Get the annotation type and the labels from the configuration file.
        with open(os.path.join(meta_dir, f'config_{name}.txt'), 'r') as f:
            contents = [i.replace("\n", "") for i in f.readlines()]
            meta['annotation_type'] = contents[1]
            meta['labels'] = contents[3].split(' ')

        # Get the image resolution and date created from the metadata file.
        with open(os.path.join(meta_dir, 'meta.json'), 'r') as f:
            contents = json.load(f)
            meta['image_size'] = contents['image_size']
            meta['generation_date'] = contents['generation_date']

        # Return the options.
        return DataFormatConverterMetadata(**meta)

    @staticmethod
    def _locate_dataset(name):
        if os.path.exists(name) and os.path.isdir(name):
            return name
        syn_path = os.path.join(os.path.expanduser('~/.agml/synthetic'), name)
        if os.path.exists(syn_path) and os.path.isdir(syn_path):
            return syn_path
        raise NotADirectoryError(
            f"The provided dataset `{name}` is not a path, "
            f"and if it were to be the name of a dataset, "
            f"then {syn_path} does not exist either.")

    def convert(self):
        """Runs the data format conversion for the input Helios dataset.

        This method runs the actual conversion of the dataset format, as specific
        for the annotation type: object detection vs. semantic segmentation. The
        newly organized dataset is written to its existing location, and if the
        conversion is successful, then the original organization is cleared out.
        """
        # If there are no annotations, the conversion is as simple as moving files.
        if self._meta.annotation_type == "none":
            try:
                self._convert_no_annotation_dataset()
            except Exception as e:
                raise e
            else:
                self._remove_existing_image_dirs()
            return

        # Otherwise, we need to reconstruct the entire structure and generate
        # the proper annotation format from the input files at the location.
        if self._meta.annotation_type == 'object_detection':
            try:
                self._convert_object_detection_dataset()
            except Exception as e:
                self._cleanup_failed_object_detection_conversion()
                raise e
            else:
                self._remove_existing_image_dirs()
        elif self._meta.annotation_type == 'semantic_segmentation':
            try:
                self._convert_semantic_segmentation_dataset()
            except Exception as e:
                self._cleanup_failed_semantic_segmentation_conversion()
                raise e
            else:
                self._remove_existing_image_dirs()
        self._make_agml_info_json()

    def _convert_no_annotation_dataset(self):
        """For datasets with no annotations, this just moves the images."""
        jpeg_images = glob.glob(
            os.path.join(self._meta.path, "image*/**/*.jpeg"), recursive = True)
        self._map_and_move_images(jpeg_images, output_dir = self._meta.path)

    def _convert_object_detection_dataset(self):
        """Converts the format of an object detection dataset to COCO JSON."""
        # Get all of the images in the dataset.
        jpeg_images = glob.glob(
            os.path.join(self._meta.path, "image*/**/*.jpeg"), recursive = True)

        # For each of the images, get their corresponding annotations.
        image_annotation_map = {}
        for image in jpeg_images:
            image_dir = os.path.dirname(image)
            image_annotation_map[image] = \
                self._convert_text_files_to_object_annotations(image_dir)

        # Create a virtual output directory structure and the new filenames.
        self._create_output_directory_structure()

        # Move all of the images.
        image_new_map = self._map_and_move_images(list(image_annotation_map.keys()))

        # Generate the image COCO JSON contents.
        image_coco, image_id_map = [], {}
        for indx, image in enumerate(image_annotation_map.keys()):
            image_id_map[image] = indx + 1
            fpath = os.path.basename(image_new_map[image])
            image_coco.append({
                'file_name': fpath, 'height': self._meta.image_size[0],
                'width': self._meta.image_size[1], 'id': indx + 1})

        # Generate the annotation COCO JSON contents.
        annotation_coco = []
        for indx, (image, annotation) in enumerate(image_annotation_map.items()):
            image_box_tracker = 1
            for label, bboxes in annotation.items():
                for box in bboxes:
                    annotation_coco.append({
                        'bbox': box.tolist(), 'area': int(box[2] * box[3]),
                        'category_id': label, 'image_id': image_id_map[image],
                        'id': image_box_tracker, 'iscrowd': 0, 'ignore': 0,
                        'segmentation': []})
                    image_box_tracker += 1

        # Create the category mapping and the meta information.
        category_coco = [
            {'name': name, 'supercategory': 'none', 'id': i}
            for i, name in enumerate(self._meta.labels)]
        info_coco = {
            'description': f"{self._meta.name}: Helios-generated dataset",
            'url': 'None', 'version': '1.0', 'year': dt.now().year,
            'contributor': 'None', 'date_created': self._meta.generation_date}

        # Save the JSON file with the annotations.
        with open(os.path.join(self._meta.path, 'annotations.json'), 'w') as f:
            json.dump({
                'images': image_coco, 'annotations': annotation_coco,
                'categories': category_coco, 'info': info_coco}, f, indent = 4)

    def _convert_text_files_to_object_annotations(self, image_dir):
        """Converts text file annotations to COCO JSON object annotations."""
        height, width = self._meta.image_size
        txt_fmt = os.path.join(image_dir, 'rectangular_labels_{0}.txt')

        # Read each of the files corresponding to the given labels.
        bboxes = {}
        for indx, label in enumerate(self._meta.labels):
            # Get the path to the specific text file and check for its existence.
            path = txt_fmt.format(label)
            if not os.path.exists(path):
                err = FileNotFoundError(
                    f"The annotation file {path} for the label `{label}` "
                    f"for the image at {image_dir} does not exist. ")
                if label == 'fruits':
                    path = txt_fmt.format('clusters')
                    if not os.path.exists(path):
                        raise err
                else:
                    raise err

            # Read the text file and get all of the lines in float format.
            with open(path, 'r') as f:
                annotations = np.array(
                    [line.replace('\n', '').strip().split(' ') for line in f.readlines()])
                annotations = annotations[:, 1:].astype(np.float32)

            # Convert the bounding boxes to COCO JSON format.
            x_c, y_c, w, h = np.rollaxis(annotations, 1)
            x_min = (x_c - w / 2) * width
            y_min = ((1 - y_c) - h / 2) * height
            w = w * width
            h = h * height
            coords = np.dstack([x_min, y_min, w, h])[0].astype(np.int32)

            # Update the bounding box dictionary.
            bboxes[indx + 1] = coords

        # Return the bounding box dictionary.
        return bboxes

    def _create_output_directory_structure(self):
        """Builds the output directory structure for the dataset."""
        if self._meta.annotation_type == 'object_detection':
            data_dir = self._meta.path
            image_dir = os.path.join(data_dir, 'images')
            os.makedirs(image_dir, exist_ok = True)
        elif self._meta.annotation_type == 'semantic_segmentation':
            data_dir = self._meta.path
            image_dir = os.path.join(data_dir, 'images')
            os.makedirs(image_dir, exist_ok = True)
            annotation_dir = os.path.join(data_dir, 'annotations')
            os.makedirs(annotation_dir, exist_ok = True)

    def _map_and_move_images(self, images, output_dir = None):
        """Maps all of the images to a new ID and moves them."""
        image_new_map = {}
        if output_dir is None:
            output_dir = os.path.join(self._meta.path, 'images')
        for image in tqdm(images, file = sys.stdout, desc = "Moving Images"):
            image_num = os.path.basename(recursive_dirname(image, 2))
            view_num = os.path.basename(recursive_dirname(image, 1))
            image_name = f"{image_num}-{view_num}.jpeg"
            shutil.copyfile(image, os.path.join(output_dir, image_name))
            image_new_map[image] = os.path.join(output_dir, image_name)
        return image_new_map

    def _cleanup_failed_object_detection_conversion(self):
        """Cleans up the remnants of a failed conversion for object detection."""
        if os.path.exists(os.path.join(self._meta.path, 'annotations.json')):
            os.remove(os.path.join(self._meta.path, 'annotations.json'))
        if os.path.exists(os.path.join(self._meta.path, 'images')):
            shutil.rmtree(os.path.join(self._meta.path, 'images'))

    def _convert_semantic_segmentation_dataset(self):
        """Converts the format of a semantic segmentation dataset to AgML's format."""
        # Get all of the images in the dataset.
        jpeg_images = glob.glob(
            os.path.join(self._meta.path, "image*/**/*.jpeg"), recursive = True)

        # For each of the images, get their corresponding annotations.
        num_to_label = None
        image_annotation_map = {}
        for image in jpeg_images:
            if num_to_label is None:
                num_to_label = self._get_number_to_label_map(
                    os.path.join(os.path.dirname(image),
                                 'semantic_segmentation_ID_mapping.txt'))
            image_annotation_map[image] = \
                self._convert_text_to_semantic_segmentation_array(
                    os.path.join(os.path.dirname(image), 'semantic_segmentation.txt'))

        # Create a virtual output directory structure and the new filenames.
        self._create_output_directory_structure()
        new_image_map = self._map_and_move_images(jpeg_images)

        # Write all of the corresponding annotations.
        annotation_dir = os.path.join(self._meta.path, 'annotations')
        for image in jpeg_images:
            new_file = os.path.join(
                annotation_dir, os.path.basename(
                    new_image_map[image].replace('.jpeg', '.png')))
            cv2.imwrite(new_file, image_annotation_map[image])

    @staticmethod
    def _get_number_to_label_map(txt):
        with open(txt, 'r') as f:
            return [{int(float(v)): k for k, v in [
                i.replace('\n', '').split(' ') for i in f.readlines()[1:]]}]

    @staticmethod
    def _convert_text_to_semantic_segmentation_array(file):
        """Converts a text file to a semantic segmentation array."""
        # Get all of the contents of the file, and convert them from strings to arrays.
        with open(file, 'r') as f:
            contents = [[i for i in line.replace('\n', '').split(' ') if i != '']
                        for line in f.readlines() if line != '']
            contents = np.array(contents).astype(np.int32)

        # Replace all extra values with zeroes.
        contents[contents == 16777215] = 0

        # Return the array.
        return contents

    def _cleanup_failed_semantic_segmentation_conversion(self):
        """Cleans up the remnants of a failed conversion for semantic segmentation."""
        if os.path.exists(os.path.join(self._meta.path, 'images')):
            shutil.rmtree(os.path.join(self._meta.path, 'images'))
        if os.path.exists(os.path.join(self._meta.path, 'annotations')):
            shutil.rmtree(os.path.join(self._meta.path, 'annotations'))

    def _remove_existing_image_dirs(self):
        """Removes the original image directories after a successful conversion."""
        image_dirs = [
            dir_ for dir_ in get_dir_list(self._meta.path) if dir_[-1].isnumeric()]
        for image_dir in image_dirs:
            shutil.rmtree(os.path.join(self._meta.path, image_dir))

    @staticmethod
    def _infer_ag_task(ml_task):
        if ml_task == 'object_detection':
            return 'fruit_detection'
        elif ml_task == 'semantic_segmentation':
            return 'fruit_segmentation'

    def _make_agml_info_json(self):
        """Constructs a JSON file with information for an AgMLDataLoader"""
        info_json = {
            'classes': self._meta.labels,
            'ml_task': self._meta.annotation_type,
            'ag_task': self._infer_ag_task(self._meta.annotation_type),
            'location': 'synthetic',
            'sensor_modality': 'rgb',
            'real_synthetic': 'synthetic',
            'n_images': len(get_file_list(os.path.join(self._meta.path, 'images'))),
            'platform': 'ground',
            'input_data_format': 'jpeg',
            'annotation_format': 'coco_json'
            if self._meta.annotation_type in ['object_detection'] else 'image'
        }

        # Save the info JSON to a file.
        with open(os.path.join(self._meta.path, '.metadata', 'agml_info.json'), 'w') as f:
            json.dump(info_json, f)






# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import abc
from functools import wraps

import cv2
import numpy as np

from agml.framework import AgMLSerializable
from agml.utils.image import imread_context
from agml.backend.tftorch import scalar_unpack


class DataObject(AgMLSerializable):
    """Stores a single piece of data and its corresponding annotation.

    This class stores an image and its corresponding annotation
    depending on the task, such as a label for a image classification
    task, a COCO JSON dictionary for an object detection task, or
    a mask for a semantic segmentation task.

    Fundamentally, it works as a two-object list, but the data is stored
    internally in a specific representation and images are only loaded
    lazily when necessary. This object is used internally in the
    `AgMLDataLoader`, and objects returned from it are returned as their
    expected contents (NumPy arrays/dictionaries/integers) when necessary.
    """
    serializable = frozenset(
        ('image_object', 'annotation_obj', 'dataset_root'))
    _abstract = frozenset(('_load_image_input', '_parse_annotation'))

    def __init__(self, image, annotation, root):
        # The `image` parameter is constant among different tasks.
        self._image_object = image

        # The `annotation` parameter varies with the task.
        self._annotation_obj = annotation

        # The `root` is the local root of the dataset. This is used
        # for object detection datasets primarily, whose COCO JSON
        # dictionary doesn't contain the full path, only the base.
        self._dataset_root = root

    def __len__(self):
        return 2

    def __getitem__(self, i):
        return self.get()[i]

    def __repr__(self):
        return f"<DataObject: {self._image_object}, {self._annotation_obj}>"

    @staticmethod
    def _parse_image(path):
        with imread_context(os.path.abspath(path)) as image:
            return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    @staticmethod
    def _parse_depth_image(path):
        with imread_context(os.path.abspath(path), flags = -1) as image:
            return image.astype(np.int32)

    @staticmethod
    def _parse_spectral_image(path):
        None # noqa, prevents `all abstract methods must be implemented`
        raise NotImplementedError("Multi/Hyperspectral images are not yet supported.")

    def get(self):
        """Returns the image and annotation pair with applied transforms.

        This method is the main exposed method to process the data. It loads
        the image and processes the annotation, then applies transformations.
        """
        return self._load()

    def _load(self):
        """Loads the image and annotation and returns them."""
        image = self._load_image_input(self._image_object)
        annotation = self._parse_annotation(self._annotation_obj)
        return image, annotation

    def __init_subclass__(cls, **kwargs):
        # Wraps derived abstract methods with the corresponding
        # docstring from this `DataObject` class and updates
        # the derived class' dictionary.
        wrapped_updates = {}
        self = super(DataObject, cls).__thisclass__ # noqa
        for name, method in cls.__dict__.items():
            if name in cls._abstract:
                wrapped = wraps(getattr(self, name))(method)
                wrapped_updates[name] = wrapped
        for name, method in wrapped_updates.items(): # noqa
            setattr(cls, name, method)

    @staticmethod
    def create(contents, task, root):
        """Creates a new `DataObject` for the corresponding task."""
        if task == 'image_classification':
            cls = ImageClassificationDataObject
        elif task == 'image_regression':
            cls = ImageRegressionDataObject
        elif task == 'object_detection':
            cls = ObjectDetectionDataObject
        elif task == 'semantic_segmentation':
            cls = SemanticSegmentationDataObject
        else:
            raise ValueError(f"Unsupported task {task}.")
        return cls(*contents, root)

    # The following methods are used to load the image and annotation.

    @abc.abstractmethod
    def _load_image_input(self, path):
        """Loads image inputs based on the task. Derived by subclasses."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _parse_annotation(self, obj):
        """Parses an annotation based on the task. Derived by subclasses."""
        raise NotImplementedError()

    # The following methods function independently whenever an
    # annotation is loaded. Depending on the task, the appropriate
    # method is called and a corresponding piece of data is returned.

    @staticmethod
    def _parse_label(obj):
        """Parses a label for an image classification task."""
        try:
            obj = int(obj)
        except TypeError:
            raise Exception(f"Could not convert object {obj} of "
                            f"type {type(obj)} to a scalar integer "
                            f"for image classification.")
        return obj

    @staticmethod
    def _parse_coco(obj):
        """Parses a COCO JSON dictionary for an object detection task."""
        annotation = {'bbox': [], 'category_id': [], 'area': [],
                      'image_id': "", 'iscrowd': [], 'segmentation': []}
        for a_set in obj:
            x, y, w, h = a_set['bbox']
            x = int(np.clip(x, 0, None))
            y = int(np.clip(y, 0, None))
            a_set['bbox'] = [x, y, w, h]
            annotation['bbox'].append(a_set['bbox'])
            annotation['category_id'].append(a_set['category_id'])
            annotation['iscrowd'].append(a_set['iscrowd'])
            annotation['segmentation'].append(a_set['segmentation'])
            annotation['area'].append(a_set['area'])
            annotation['image_id'] = a_set['image_id']
        for key, value in annotation.items():
            # Creating nested sequences from ragged arrays (see numpy).
            if key in ['segmentation']:
                out = np.array(value, dtype = object)
            else:
                out = np.array(value)
                if np.isscalar(out):
                    out = scalar_unpack(out)
            annotation[key] = out
        return annotation

    @staticmethod
    def _parse_mask(obj):
        """Parses a mask for a semantic segmentation task."""
        with imread_context(os.path.realpath(obj)) as image:
            if image.ndim == 3:
                if not np.all(image[:, :, 0] == image[:, :, 1]):
                    raise TypeError(
                        f"Invalid annotation mask of shape {image.shape}.")
                image = image[:, :, 0]
            return np.squeeze(image)


class ImageClassificationDataObject(DataObject):
    """Serves as a `DataObject` for image classification tasks."""
    def _load_image_input(self, path):
        return self._parse_image(path)

    def _parse_annotation(self, obj):
        return self._parse_label(obj)


class ImageRegressionDataObject(DataObject):
    """Serves as a `DataObject` for image regression tasks."""
    def _load_image_input(self, contents):
        # The easy case, when there is only one input image.
        if isinstance(contents, str) and os.path.exists(contents):
            return self._parse_image(contents)

        # Otherwise, we have a dictionary containing multiple
        # input types, so we need to independently load those.
        images = dict.fromkeys(contents.keys(), None)
        for c_type, path in contents.items():
            if c_type == 'image':
                images[c_type] = self._parse_image(path)
            elif c_type == 'depth_image':
                images[c_type] = self._parse_depth_image(path)
            else:
                images[c_type] = self._parse_spectral_image(path)
        return images

    def _parse_annotation(self, obj):
        return obj


class ObjectDetectionDataObject(DataObject):
    """Serves as a `DataObject` from object detection tasks."""
    def _load_image_input(self, path):
        path = os.path.join(self._dataset_root, 'images', path)
        return self._parse_image(path)

    def _parse_annotation(self, obj):
        return self._parse_coco(obj)


class SemanticSegmentationDataObject(DataObject):
    """Serves as a `DataObject` for semantic segmentation tasks."""
    def _load_image_input(self, path):
        return self._parse_image(path)

    def _parse_annotation(self, obj):
        return self._parse_mask(obj)






# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import numpy as np

import matplotlib.pyplot as plt

from agml.utils.general import resolve_tuple_values
from agml.viz.tools import (
    format_image, get_colormap, auto_resolve_image, show_when_allowed
)


def _reduce_categorical_mask(mask):
    """Reduces a categorical mask label to one-dimensional."""
    return np.argmax(mask, axis = -1)


def _preprocess_mask(mask):
    """Preprocesses a mask with a distinct colorscheme."""
    if mask.ndim == 3:
        if not np.all(mask[:, :, 0] == mask[:, :, 1]):
            mask = _reduce_categorical_mask(mask)
    if mask.ndim == 2:
        mask = np.dstack((mask, mask, mask))

    # Add a colorscheme to the labels.
    label_indx = 1
    while True:
        x_values, y_values, *_ = np.where(mask == label_indx)
        coords = np.stack([x_values, y_values]).T
        if coords.size == 0:
            break
        mask[tuple([*coords.T])] = get_colormap()[label_indx]
        label_indx += 1
    return mask


def _mask_2d_to_3d(mask):
    """Converts a mask from 2-dimensional to channel-by-channel."""
    if mask.ndim == 3:
        return mask
    mask = np.squeeze(mask)
    channels = np.unique(mask)[1:]
    if len(channels) == 0:
        return np.zeros(shape = (*mask.shape[:2], 1))
    out = np.zeros(shape = (*mask.shape[:2], int(max(channels))))
    iter_idx = 1
    while True:
        if iter_idx > max(channels):
            break
        elif iter_idx not in channels:
            out[:, :, iter_idx - 1] = np.zeros(shape = (mask.shape[:2]))
            iter_idx += 1
            continue
        coords = np.stack(np.where(mask == iter_idx)[:2]).T
        channel = np.zeros(shape = (*mask.shape[:2],))
        channel[tuple([*coords.T])] = 1
        out[:, :, iter_idx - 1] = channel
        iter_idx += 1
    return out


def output_to_mask(mask):
    """Converts an output annotation mask into a visual segmentation mask.

    Given the output segmentation mask from a model (a 2-dimensional
    image array with a number representing a class label) this method
    will convert it into an image with colors for the regions with
    different labels. This is a purely visual transformation.

    Parameters
    ----------
    mask : np.ndarray
        The above cases of a possible segmentation mask.

    Returns
    -------
    The colorful visually formatted mask.
    """
    return _preprocess_mask(mask)


@show_when_allowed
@auto_resolve_image
def visualize_image_and_mask(image, mask = None):
    """Visualizes an image and its corresponding segmentation mask.

    Creates a 1 row, 2 column frame and displays an image with
    its corresponding segmentation mask. Applies a distinct
    colorscheme to the mask to visualize colors more clearly.

    Parameters
    ----------
    image : Any
        Either the original image, or a tuple containing the image
        and its mask (if using a DataLoader, for example).
    mask : np.ndarray
        The output mask.

    Returns
    -------
    The matplotlib figure with the images.
    """
    image, mask = resolve_tuple_values(
        image, mask, custom_error =
        "If `image` is a tuple/list, it should contain "
        "two values: the image and its mask.")
    image = format_image(image)
    mask = _preprocess_mask(format_image(mask, mask = True))

    fig, axes = plt.subplots(1, 2, figsize = (10, 5))
    axes[0].imshow(image)
    axes[1].imshow(mask)
    for ax in axes:
        ax.set_aspect('equal')
        ax.axis('off')
    return fig


@show_when_allowed
@auto_resolve_image
def visualize_image_mask_and_predicted(image, mask = None, predicted = None):
    """Visualizes an image, its segmentation mask, and a predicted mask.

    Creates a 1 row, 3 column frame and displays an image with
    its corresponding segmentation mask and a predicted segmentation
    mask as given in `predicted`. This figure includes axis titles
    for the image, mask, and predicted mask, for differentiation.

    Parameters
    ----------
    image : Any
        Either the original image, or a tuple containing the image
        and its mask (if using a DataLoader, for example).
    mask : np.ndarray
        The output mask.
    predicted : np.ndarray
        The predicted mask.

    Returns
    -------
    The matplotlib figure with the images.
    """
    image, mask, predicted = resolve_tuple_values(
        image, mask, predicted, custom_error =
        "If `image` is a tuple/list, it should contain three "
        "values: the image, its mask, and a predicted mask.")
    image = format_image(image)
    mask = _preprocess_mask(format_image(mask))
    predicted = _preprocess_mask(format_image(predicted))

    fig, axes = plt.subplots(1, 3, figsize = (10, 5))
    for im, ax, label in zip(
        [image, mask, predicted], axes,
        ['Input Image', 'Ground Truth Mask', 'Predicted Mask']
    ):
        ax.imshow(im)
        ax.set_title(label)
        ax.set_aspect('equal')
        ax.axis('off')
    fig.tight_layout()
    return fig


@auto_resolve_image
def overlay_segmentation_masks(image, mask = None, border = True):
    """Overlays segmentation masks over an image.

    Creates a single image and displays the segmentation annotations
    from the mask overlaid onto the main image. The mask will be
    semi-transparent and have a colorscheme applied to it in order
    to facilitate easier visualization of it.

    Parameters
    ----------
    image : Any
        Either the original image, or a tuple containing the image
        and its mask (if using a DataLoader, for example).
    mask : np.ndarray
        The output mask (should be an array with dimension 2).
    border : bool
        Whether to add a border to the segmentation overlay.

    Returns
    -------
    A `np.ndarray` representing the image.
    """
    image, mask = resolve_tuple_values(
        image, mask, custom_error =
        "If `image` is a tuple/list, it should contain "
        "two values: the image and its mask.")
    image = format_image(image)
    mask = _mask_2d_to_3d(format_image(mask))

    # Plot the segmentation masks over the image
    mask = np.transpose(mask, (2, 0, 1))
    for level, label in enumerate(mask):
        label = np.expand_dims(label, axis = -1)
        label = (label * (255 / (np.max(label, axis = (0, 1))))) \
                .astype(np.uint8)

        # Find contours and plot them
        _, thresh = cv2.threshold(label, 127, 255, cv2.THRESH_BINARY)
        contours, _ = cv2.findContours(
            thresh, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
        for indx, contour in enumerate(contours):
            color = get_colormap()[level + 1]
            overlay = image.copy()
            cv2.fillPoly(overlay, pts = [contour], color = color)
            label = label.astype(np.float32)
            image = cv2.addWeighted(overlay, 0.3, image, 0.7, 0)
            if border:
                image = cv2.polylines(
                    image, pts = [contour], isClosed = True,
                    color = color, thickness = 2)
    return image


@show_when_allowed
@auto_resolve_image
def visualize_overlaid_masks(image, mask = None, border = True):
    """Displays an image with segmentation masks overlaid on it.

    See `overlay_segmentation_masks` for an explanation of procedure.

    Parameters
    ----------
    image : Any
        Either the original image, or a tuple containing the image
        and its mask (if using a DataLoader, for example).
    mask : np.ndarray
        The output mask (should be an array with dimension 2).
    border : bool
        Whether to add a border to the segmentation overlay.

    Returns
    -------
    A `np.ndarray` representing the image.
    """
    # Plot the segmentation masks over the image
    image = overlay_segmentation_masks(image, mask, border)

    # Plot the figure
    plt.figure(figsize = (10, 10))
    plt.imshow(image)
    plt.gca().axis('off')
    plt.gca().set_aspect('equal')
    return image








# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A set of data preprocessing functions for `AgMLDataLoaders`.
"""

import inspect

import torch
import numpy as np
from PIL import Image
import albumentations as A
from albumentations.pytorch import ToTensorV2


class EfficientDetPreprocessor(object):
    """A preprocessor which prepares a data sample for `EfficientDet`.
    
    This class can be used to construct a preprocessing pipeline which
    auto-formats the data in an `AgMLDataLoader` for object detection into
    the format necessary for training an `EfficientDet` model. By default,
    this includes resizing images, converting bounding boxes to `yxyx`, and
    finally preparing the image and annotation for PyTorch.
    
    Using this class supersedes the need for any other transformations or
    even image resizing. It can be used as follows:
    
    > loader = agml.data.AgMLDataLoader('grape_detection_californiaday')
    > processor = agml.models.EfficientDetPreprocessor(
    >     image_size = 512, augmentation = [A.HorizontalFlip(0.5)])
    > loader.transform(dual_transform = processor)
    
    Parameters
    ----------
    image_size : int
        The size to which images will be resized (default of 512).
    augmentation : Any
        Either a list of albumentations transforms (without being wrapped
        into a Compose object), or a custom method which accepts three
        arguments: `image`, `bboxes`, and `labels`, and returns the same.
        The `bboxes` will be in the XYXY format.
    
    Notes
    -----
    - Passing `augmentation = None`, the default, is equivalent to preparing
      a validation or test loader: preprocessing is applied, but no transforms.
    - Note that if you pass a custom augmentation method, the resulting output
      is expected to be in PyTorch's format (image should be a tensor with its
      first dimension being the image's channels, for example).
    """

    def __init__(self, image_size = 512, augmentation = None):
        # Parse the image size.
        if isinstance(image_size, int):
            image_size = (image_size, image_size)
        elif isinstance(image_size, (list, tuple)):
            if not len(image_size) == 2:
                raise ValueError(
                    "Expected either an integer or sequence of 2 values "
                    "for `image_size`, instead got ({image_size}).")

        # Construct the applied input augmentation.
        self._albumentations_params = dict(
            bbox_params = A.BboxParams(
                format = "pascal_voc", min_area = 0,
                min_visibility = 0, label_fields = ["labels"]),
            standard_augmentations = [
                A.Resize(height = image_size[0],
                         width = image_size[1], p = 1),
                ToTensorV2(p = 1)])
        self._check_and_make_augmentation(augmentation)

    def _check_and_make_augmentation(self, augmentation):
        """Constructs the applied augmentation from the inputs."""
        # If no augmentation is provided, then use the defaults.
        if augmentation is None:
            self._augmentation = A.Compose([
                *self._albumentations_params['standard_augmentations']
            ], p = 1.0, bbox_params = self._albumentations_params['bbox_params'])

        # If a list of augmentations are provided, then use those augmentations
        # wrapped alongside the default ones in a `Compose` object.
        elif isinstance(augmentation, list):
            if not all(isinstance(a, A.BasicTransform) for a in augmentation):
                raise ValueError(
                    "If providing a list of transforms, all of them must be "
                    f"albumentations augmentations, instead got {augmentation} "
                    f"of types {[type(i) for i in augmentation]}.")
            self._augmentation = A.Compose([
                *augmentation, *self._albumentations_params['standard_augmentations']
            ], p = 1.0, bbox_params = self._albumentations_params['bbox_params'])

        # Otherwise, the augmentation should be a method with three input
        # arguments, so check it and then wrap it into an application method.
        else:
            if not len(inspect.signature(augmentation).parameters):
                raise ValueError(
                    f"The input augmentation should have three input arguments, "
                    f"instead got {inspect.signature(augmentation).parameters}.")
            self._method_augmentation = augmentation
            self._augmentation = self._apply_method_augmentation

    def _apply_method_augmentation(self, image, bboxes, labels):
        # Wrapper method to apply a user-provided method augmentation.
        image, bboxes, labels = self._method_augmentation(image, bboxes, labels)
        return {'image': image, 'bboxes': bboxes, 'labels': labels}

    def __call__(self, image, annotation):
        # Convert the image type.
        image = Image.fromarray(image)

        # Clip the bounding boxes to the image shape to prevent errors.
        bboxes = np.array(annotation['bbox']).astype(np.int32)
        x_min = bboxes[:, 0]
        y_min = bboxes[:, 1]
        x_max = bboxes[:, 2] + x_min
        y_max = bboxes[:, 3] + y_min
        x_min, y_min = np.clip(x_min, 0, image.width), \
                       np.clip(y_min, 0, image.height)
        x_max, y_max = np.clip(x_max, 0, image.width), \
                       np.clip(y_max, 0, image.height)

        # Reconstruct the boxes and get the class labels.
        bboxes = np.dstack((x_min, y_min, x_max, y_max)).squeeze(axis = 0)
        class_labels = np.array(annotation['category_id']).squeeze()

        # Add an extra dimension to labels for consistency.
        if class_labels.ndim == 0:
            class_labels = np.expand_dims(class_labels, axis = 0)

        # Add an albumentations augmentation.
        sample = {'image': np.array(image, dtype = np.float32),
                  'bboxes': bboxes, 'labels': class_labels}
        sample = self._augmentation(**sample)
        image = sample['image']
        bboxes = np.array(sample['bboxes'])
        labels = sample['labels']

        # Convert 1-channel and 4-channel to 3-channel.
        if image.shape[0] == 1:
            image = torch.tile(image, (3, 1, 1))
        if image.shape[0] == 4:
            image = image[:3]

        # Convert to yxyx from xyxy.
        _, new_h, new_w = image.shape
        if bboxes.ndim == 1:
            bboxes = np.expand_dims(bboxes, axis = 0)
        bboxes[:, [0, 1, 2, 3]] = bboxes[:, [1, 0, 3, 2]]

        # Create the target from the annotations.
        target = {
            "bboxes": torch.as_tensor(
                bboxes, dtype = torch.float32),
            "labels": torch.as_tensor(labels),
            "img_size": torch.tensor([new_h, new_w]),
            "img_scale": torch.tensor([1.0])}
        return image, target





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Controls the AgML backend system.

The first part of the backend is the backend deep learning library.
The backend, one of {TensorFlow, PyTorch}, primarily exists for internal
purposes, e.g. figuring out which methods to use in the data module or
the actual model configuring/training modules.

Secondly, the backend controls the loading/saving procedure for files within
AgML, specifically data loaded or generated from the data module.
"""

from .config import (
    data_save_path,
    set_data_save_path,
    synthetic_data_save_path,
    set_synthetic_save_path,
    model_save_path,
    set_model_save_path,
    clear_all_datasets,
    downloaded_datasets
)
from .tftorch import get_backend, set_backend
from .random import set_seed
from . import experimental



# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import shutil
import logging

from agml.utils.logging import log


# The super base save directory for AgML. This is the overriding base
# save directory and default original save directory. This is saved here
# because the default path to save datasets to can be overridden by
# `set_dataset_save_path()`, but we need to keep the super directory to
# access the config file which stores this information in the first place.
SUPER_BASE_DIR = os.path.join(os.path.expanduser('~'), '.agml')


# This is the path to the saved datasets. By default, this saves to
# SUPER_BASE_DIR/datasets, but can be overridden. The value of this
# is set upon the instantiation of the module (see below code).
DATASET_SAVE_DIR: str



# This is similar to `DATASET_SAVE_DIR`, but is for synthetically generated
# datasets using Helios. By default, this will be SUPER_BASE_DIR/synthetic,
# but it can be overridden. The value is set upon instantiation of the module.
SYNTHETIC_SAVE_DIR: str


# This is the path to any downloaded models. By default, this saves to
# SUPER_BASE_DIR/models, but can be overridden. The value is set upon
# instantiation of the module (see the below code).
MODEL_SAVE_DIR: str


# Loads the configuration info. We don't cache this since it may
# change if the user decides to change the path. This method
# also runs upon the first import of AgML to set it properly.
def _load_config_info():
    global DATASET_SAVE_DIR, SYNTHETIC_SAVE_DIR, MODEL_SAVE_DIR
    try:
        with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'r') as f:
            contents = json.load(f)
            DATASET_SAVE_DIR = contents['data_path']
            SYNTHETIC_SAVE_DIR = contents['synthetic_data_path']
            MODEL_SAVE_DIR = contents['model_path']
    except (OSError, KeyError):
        with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'w') as f:
            json.dump({'data_path': os.path.join(SUPER_BASE_DIR, 'datasets'),
                       'synthetic_data_path': os.path.join(
                           SUPER_BASE_DIR, 'synthetic'),
                       'model_path': os.path.join(SUPER_BASE_DIR, 'models')}, f)
        _load_config_info()
_load_config_info()


def data_save_path():
    """Returns the default dataset save path for AgML."""
    global DATASET_SAVE_DIR
    return DATASET_SAVE_DIR


def set_data_save_path(location = None):
    """Sets the default dataset save path for AgML.
    Changing the data save path using this method permanently changes
    the data save path for all future sessions, until it is changed
    or switched back to the original. If you just want to download one
    dataset to a different path, use the `dataset_path` argument.
    Parameters
    ----------
    location : str
        The location to save the data to.
    Returns
    -------
    The fully expanded location.
    """
    global SUPER_BASE_DIR
    if location is None or location == 'reset':
        location = os.path.join(SUPER_BASE_DIR, 'datasets')
    location = os.path.expanduser(location)
    if not os.path.exists(location) and not os.path.isdir(location):
        raise NotADirectoryError(
            f"The provided destination {location} does "
            f"not exist, or is not a directory.")
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'r') as f:
        contents = json.load(f)
    contents['data_path'] = os.path.realpath(os.path.abspath(location))
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'w') as f:
        json.dump(contents, f)
    return


def synthetic_data_save_path():
    """Returns the default synthetic data save path for AgML."""
    global SYNTHETIC_SAVE_DIR
    return SYNTHETIC_SAVE_DIR


def set_synthetic_save_path(location = None):
    """Sets the default synthetic data save path for AgML.
    Changing the data save path using this method permanently changes
    the data save path for all future sessions, until it is changed
    or switched back to the original. If you just want to download one
    dataset to a different path, use the `dataset_path` argument.
    Parameters
    ----------
    location : str
        The location to save the data to.
    Returns
    -------
    The fully expanded location.
    """
    global SUPER_BASE_DIR
    if location is None or location == 'reset':
        location = os.path.join(SUPER_BASE_DIR, 'synthetic')
    location = os.path.expanduser(location)
    if not os.path.exists(location) and not os.path.isdir(location):
        raise NotADirectoryError(
            f"The provided destination {location} does "
            f"not exist, or is not a directory.")
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'r') as f:
        contents = json.load(f)
    contents['synthetic_data_path'] = os.path.realpath(os.path.abspath(location))
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'w') as f:
        json.dump(contents, f)
    return


def model_save_path():
    """Returns the default model save path for AgML."""
    global MODEL_SAVE_DIR
    return MODEL_SAVE_DIR


def set_model_save_path(location = None):
    """Sets the default model save path for AgML.
    Changing the data save path using this method permanently changes
    the data save path for all future sessions, until it is changed
    or switched back to the original.
    Parameters
    ----------
    location : str
        The location to save the model to.
    Returns
    -------
    The fully expanded location.
    """
    global SUPER_BASE_DIR
    if location is None or location == 'reset':
        location = os.path.join(SUPER_BASE_DIR, 'models')
    location = os.path.expanduser(location)
    if not os.path.exists(location) and not os.path.isdir(location):
        raise NotADirectoryError(
            f"The provided destination {location} does "
            f"not exist, or is not a directory.")
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'r') as f:
        contents = json.load(f)
    contents['model_path'] = os.path.realpath(os.path.abspath(location))
    with open(os.path.join(SUPER_BASE_DIR, 'config.json'), 'w') as f:
        json.dump(contents, f)
    return



def clear_all_datasets():
    """Deletes all of the datasets within the AgML local storage."""
    log("Entering AgML interactive dataset deletion mode.", logging.WARNING)
    msg_format = '\033[91m{0}\033[0m'
    if not input(msg_format.format(
            "Please confirm that you want to delete datasets: [y|n] ")) == "y":
        print("Aborting dataset deletion.")
    local_datasets = os.listdir(data_save_path())
    deleted_datasets = []
    for dataset in local_datasets:
        if not os.path.isdir(os.path.join(
                data_save_path(), dataset)):
            continue
        if not input(msg_format.format(
                f"Delete dataset '{dataset}'? [y|n] ")) == "y":
            continue
        shutil.rmtree(os.path.join(data_save_path(), dataset))
        deleted_datasets.append(dataset)
    print(f"Deleted datasets {deleted_datasets}.")
    log("Exiting AgML interactive dataset deletion mode.", logging.WARNING)


def downloaded_datasets():
    """Lists downloaded datasets in ~/.agml/datasets"""
    return [d for d in os.listdir(
        data_save_path()) if os.path.isdir(
        os.path.join(data_save_path(), d))]


# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import numpy as np
import matplotlib.pyplot as plt

from agml.utils.general import resolve_tuple_values
from agml.backend.tftorch import as_scalar, scalar_unpack, is_array_like
from agml.data.tools import (
    _resolve_coco_annotations, convert_bbox_format # noqa
)
from agml.viz.tools import (
    get_colormap, auto_resolve_image, format_image, show_when_allowed
)


def _resolve_proportional_bboxes(coords, shape):
    """Resolves float vs. integer bounding boxes."""
    coords = scalar_unpack(coords)
    if all(isinstance(i, float) for i in coords):
        if coords[0] <= 1:
            x, y, width, height = coords
            y *= shape[0]; height *= shape[0]  # noqa
            x *= shape[1]; width *= shape[1]  # noqa
            coords = [x, y, width, height]
        return [int(round(c)) for c in coords]
    elif all(isinstance(i, int) for i in coords):
        return coords
    raise TypeError(
        f"Got multiple types for coordinates: "
        f"{[type(i) for i in coords]}.")


@auto_resolve_image
def annotate_bboxes_on_image(
        image, bboxes = None, labels = None, bbox_format = None):
    """Annotates bounding boxes onto an image.

    Given an image with bounding boxes and labels, this method will
    annotate the bounding boxes directly onto the image (with category
    label). Bounding boxes are expected to be in COCO JSON format, that
    is, (`x_min`, `y_max`, `width`, `height`). Use the helper method
    `agml.data.convert_bboxes` to format your bounding boxes as such.

    Parameters
    ----------
    image : Any
        Either the image, or a tuple consisting of the image,
        bounding boxes, and (optional) labels.
    bboxes : Any
        The bounding boxes in COCO JSON format. This can be either
        a dictionary with COCO JSON annotations, or just the boxes.
    labels : Any
        Optional category labels for the bounding box color.
    bbox_format : optional
        The format of the bounding box (for non-COCO-JSON bounding
        boxes).  See `agml.data.convert_bbox_format` for information
        on the valid parameters for the format.

    Returns
    -------
    The annotated image.
    """
    image, bboxes, labels = resolve_tuple_values(
        image, bboxes, labels, custom_error =
        "If `image` is a tuple/list, it should contain "
        "three values: the image, mask, and (optionally) labels.")
    if isinstance(bboxes, dict):
        try:
            bboxes = _resolve_coco_annotations(bboxes)['bboxes']
        except KeyError:
            bboxes = _resolve_coco_annotations(bboxes)['bbox']
    if len(bboxes) == 0:
        return image

    # This ensures that we don't get a resolution error with cv2.
    image = np.ascontiguousarray(format_image(image))
    if labels is None:
        labels = [1] * len(bboxes)
    if bbox_format is not None:
        bboxes = convert_bbox_format(bboxes, bbox_format)
    if not isinstance(bboxes[0], (list, np.ndarray)):
        if not is_array_like(bboxes[0]):
            bboxes = [bboxes]

    for bbox, label in zip(bboxes, labels):
        bbox = scalar_unpack(bbox)
        bbox = [int(i) for i in bbox]
        x1, y1, width, height = \
            _resolve_proportional_bboxes(bbox, image.shape)
        x2, y2 = x1 + width, y1 + height
        cv2.rectangle(image, (x1, y1), (x2, y2),
                      get_colormap()[as_scalar(label)], 2)
    return image


@show_when_allowed
@auto_resolve_image
def visualize_image_and_boxes(
        image, bboxes = None, labels = None, bbox_format = None):
    """Visualizes an image with annotated bounding boxes.

    This method performs the same actions as `annotate_bboxes_on_image`,
    but simply displays the image once it has been formatted.

    Parameters
    ----------
    image : Any
        Either the image, or a tuple consisting of the image,
        bounding boxes, and (optional) labels.
    bboxes : Any
        The bounding boxes in COCO JSON format.
    labels : Any
        Optional category labels for the bounding box color.
    bbox_format : optional
        The format of the bounding box (for non-COCO-JSON bounding
        boxes).  See `agml.data.convert_bbox_format` for information
        on the valid parameters for the format.

    Returns
    -------
    The matplotlib figure with the image.
    """
    image, bboxes, labels = resolve_tuple_values(
        image, bboxes, labels, custom_error =
        "If `image` is a tuple/list, it should contain "
        "three values: the image, boxes, and (optionally) labels.")
    image = annotate_bboxes_on_image(image, bboxes, labels, bbox_format)

    plt.figure(figsize = (10, 10))
    plt.imshow(image)
    plt.gca().axis('off')
    plt.gca().set_aspect('equal')
    return plt.gcf()


@show_when_allowed
@auto_resolve_image
def visualize_real_and_predicted_bboxes(
        image, truth_boxes = None, predicted_boxes = None):
    """Visualizes an image with annotated truth/predicted bounding boxes.

    This method performs the same actions as `annotate_bboxes_on_image`,
    but simply displays the image once it has been formatted. It displays
    the input predicted and ground truth boxes side-by-side.

    Parameters
    ----------
    image : Any
        Either the image, or a tuple consisting of the image,
        bounding boxes, and (optional) labels.
    truth_boxes : Any
        The true bounding boxes in COCO JSON format.
    predicted_boxes : Any
        The predicted bounding boxes in COCO JSON format.

    Returns
    -------
    The matplotlib figure with the image.
    """
    image, truth_boxes, predicted_boxes = resolve_tuple_values(
        image, truth_boxes, predicted_boxes, custom_error =
        "If `image` is a tuple/list, it should contain "
        "three values: the image, truth boxes, and predicted boxes.")
    real_image = annotate_bboxes_on_image(image.copy(), truth_boxes)
    predicted_image = annotate_bboxes_on_image(image.copy(), predicted_boxes)

    # Create two side-by-side figures with the images.
    fig, axes = plt.subplots(1, 2, figsize = (10, 5))
    for ax, img, label in zip(
            axes, (real_image, predicted_image),
            ("Ground Truth Boxes", "Predicted Boxes")):
        ax.imshow(img)
        ax.set_axis_off()
        ax.set_title(label, fontsize = 15)
        ax.set_aspect('equal')
    fig.tight_layout()
    return fig


@show_when_allowed
@auto_resolve_image
def visualize_image_and_many_boxes(image, *boxes, titles = None):
    """Visualizes an image with an arbitrary number of bounding boxes.

    This method performs the same actions as `annotate_bboxes_on_image`,
    but simply displays the image once it has been formatted.

    Parameters
    ----------
    image : Any
        Either the image, or a tuple consisting of the image,
        bounding boxes, and (optional) labels.
    boxes : Any
        An arbitrary number of bounding boxes.
    titles : Any
        The titles for the axes.

    Returns
    -------
    The matplotlib figure with the image.
    """
    images = [annotate_bboxes_on_image(image.copy(), box) for box in boxes]
    titles = titles if titles is not None else [None] * len(images)

    # Create two side-by-side figures with the images.
    fig, axes = plt.subplots(1, len(images), figsize = (len(images) * 8, 8))
    for ax, img, title in zip(axes, images, titles):
        ax.imshow(img)
        ax.set_axis_off()
        if title is not None:
            ax.set_title(title, fontsize = 15)
        ax.set_aspect('equal')
    fig.tight_layout()
    return fig




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import numpy as np


class seed_context(object):
    """Creates a context with a custom random seed, then resets it.

    This allows for setting a custom seed (for reproducibility) in a
    context, then resetting the original state after exiting, to
    prevent interfering with the program execution.
    """

    def __init__(self, seed):
        self._seed = seed
        self._prev_state = None

    def __enter__(self):
        self._prev_state = np.random.get_state()
        np.random.seed(self._seed)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        np.random.set_state(self._prev_state)
        self._prev_state = None

    def reset(self):
        np.random.seed(self._seed)


_RANDOM_STATE_DOC = """random_state : int
\t    An integer representing a random seed to run the method using."""


def inject_random_state(f):
    """Runs a method inside of a context if a random seed is provided."""
    if "{random_state}" in f.__doc__:
        f.__doc__ = f.__doc__.format(random_state = _RANDOM_STATE_DOC)
    else:
        f.__doc__ += _RANDOM_STATE_DOC

    @functools.wraps(f)
    def _run(*args, random_state = None, **kwargs):
        # If no random state is provided, run the method as normal.
        if random_state is None:
            return f(*args, **kwargs)

        # Otherwise, run the method in a context with the random seed.
        if not isinstance(random_state, int):
            raise TypeError(
                f"Expected `random_state` argument to be "
                f"an integer, instead got ({random_state}) "
                f"of type ({type(random_state)}).")

        with seed_context(random_state):
            result = f(*args, **kwargs)
        return result
    return _run




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for working with synthetic data in the Helios format."""

import os

import math
from typing import List, Union, Tuple

import numpy as np
from scipy.signal import sawtooth

from agml.utils.io import get_dir_list


def generate_environment_map(
        origin: List[Union[int, float]],
        plant_spacing: int = 1,
        row_spacing: int = 4,
        trees_per_row: int = 1,
        rows: int = 1,
        plant_height: int = 1) -> list:
    """Generates an environment map in the Helios format.

    This method returns a list of NumPy arrays which contain coordinates
    of the canopies to be generated as specified by the provided parameters.
    These can be plotted/inspected to view the arrangement of the plants.

    Parameters
    ----------
    origin: {list, np,ndarray}
        A three-dimensional coordinate indicating the origin of the plot.
    plant_spacing : int
        The spacing of the plants.
    row_spacing : int
        The spacing between each row of plants.
    trees_per_row : int
        The number of trees to keep in each row of plants.
    rows : int
        The number of rows to generate.
    plant_height : 1
        The height of each plant generated.

    Returns
    -------
    A list of NumPy arrays representing each row, with the contents of each
    array being the coordinates of each plant.
    """
    # If there is only one row and only one tree in the row, then there
    # is only one plant whose coordinates needs to be generated.
    if trees_per_row == 1 and rows == 1:
        return [[np.concatenate([np.array([origin[0]]), np.array([origin[1]]),
                                 plant_height / 2 + np.array([origin[2]])])]]

    # Calculate the x- positions for the trees.
    if trees_per_row % 2 != 0 and trees_per_row > 1:  # odd number
        x_pos = np.concatenate(
            [np.array([origin[0]]), np.linspace(
                origin[0] + plant_spacing, trees_per_row / 2 * plant_spacing -
                plant_spacing / 2, int(trees_per_row / 2))])
        x_pos = np.concatenate([np.sort(-x_pos[1:]), x_pos])
    else:  # even number
        x_pos = np.linspace(origin[0] + (plant_spacing / 2),
                            trees_per_row / 2 * plant_spacing - plant_spacing / 2,
                            int(trees_per_row / 2))
        x_pos = np.concatenate([np.sort(-x_pos[:]), x_pos])

    # Calculate the y- positions for the trees.
    if rows % 2 != 0:  # odd number
        y_pos = np.concatenate(
            [np.array([origin[0]]), np.linspace(
                origin[1] + row_spacing, rows / 2 * row_spacing -
                row_spacing / 2, int(rows / 2))])
        y_pos = np.concatenate([np.sort(-y_pos[1:]), y_pos])
    else:
        y_pos = np.linspace(origin[1] + (row_spacing / 2),
                            rows / 2 * row_spacing - row_spacing / 2, int(rows / 2))
        y_pos = np.concatenate([np.sort(-y_pos[:]), y_pos])

    # Return the x- and y- positions.
    return [[[x_pos[x], y_pos[y], plant_height / 2 + origin[2]]
             for x in range(len(x_pos))] for y in range(len(y_pos))]


def generate_camera_positions(
        camera_type: str,
        num_views: int,
        origin: List[Union[int, float]] = None,
        camera_spacing: int = 2,
        crop_distance: int = 4,
        height: int = 1) -> Tuple:
    """Generates camera placement and lookat positions in the Helios format.

    This method, given the origin and a number of camera positions to generate
    (as specified in `num_views`), returns the 3-d coordinates of the camera
    positions as well as lookat vectors symbolizing their perspective of the
    provided origin point.

    Parameters
    ----------
    camera_type : str
        The type of camera position to generate: 'circular', 'linear',
        or 'aerial'. The lookat vectors are adjusted accordingly.
    num_views : int
        The number of different camera view positions to generate.
    origin : {list, np.ndarray}
        The origin point the cameras are focused on.
    camera_spacing : int
        The spacing of the cameras around a central point.
    crop_distance : int
        The distance of the cameras from the crop itself.
    height : int
        The height of the cameras relative to the crop.

    Returns
    -------
    A tuple of two values: the camera positions and the lookat vectors.
    """
    if origin is None:
        origin = [0, 0, 0]

    if camera_type == 'circular':
        return [[math.cos(2 * math.pi / num_views * x) * crop_distance, 
                  math.sin(2 * math.pi / num_views * x) * crop_distance, height] 
                 for x in range(0, num_views)], [[0, 0, 1] for _ in range(0, num_views)]

    elif camera_type == 'linear':
        camera_pos = np.arange(
            origin[0], camera_spacing * num_views + origin[0], camera_spacing)
        return [[camera_pos[x], crop_distance + origin[1], height]
                for x in range(len(camera_pos))], \
                [[camera_pos[x], origin[1], height] for x in range(len(camera_pos))]

    elif camera_type == 'aerial':
        t = camera_spacing * np.linspace(0, 1, num_views)
        triangle = camera_spacing * sawtooth(1 * np.pi * 5 * t, 0.5)
        return [[t[x] + origin[0], triangle[x] + origin[1],
                  crop_distance + origin[2]] for x in range(len(triangle))], \
                [[t[x] + origin[0], triangle[x] + origin[1] + 1,
                  height + origin[2]] for x in range(len(triangle))]

    else:
        raise ValueError(f"Got `camera_type`: ({camera_type}), "
                         f"expected either `circular`, `linear`, or `aerial`.")


def _is_agml_converted(dataset_path):
    """Returns whether a Helios dataset has been converted to AgML format."""
    return os.path.exists(os.path.join(dataset_path, '.metadata', 'agml_info.json')) \
            or get_dir_list(dataset_path) == ['.metadata'] # dataset with no annotations



# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import cv2
import glob
import math
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from math import pi, floor
from scipy import signal

import numpy as np


# Represents an empty object, but allows passing `None`
# as an independent object in certain cases.
NoArgument = object()


def placeholder(obj):
    """Equivalent of lambda x: x, but enables pickling."""
    return obj


def to_camel_case(s):
    """Converts a given string `s` to camel case."""
    s = re.sub(r"(_|-)+", " ", s).title().replace(" ", "") # noqa
    return ''.join(s)


def resolve_list_value(val):
    """Determines whether a list contains one or multiple values."""
    if len(val) == 1:
        return val[0]
    return val


def resolve_tuple_values(*inputs, custom_error = None):
    """Determines whether `inputs[0]` contains two values or
    they are distributed amongst the values in `inputs`. """
    if isinstance(inputs[0], (list, tuple)) and all(c is None for c in inputs[1:]):
        if len(inputs[0]) != len(inputs):
            # special case for COCO JSON
            if len(inputs) == 3 and len(inputs[0]) == 2 and isinstance(inputs[0][1], dict):
                try:
                    return inputs[0][0], inputs[0][1]['bbox'], inputs[0][1]['category_id']
                except KeyError:
                    return inputs[0][0], inputs[0][1]['bboxes'], inputs[0][1]['labels']
            if custom_error is not None:
                raise ValueError(custom_error)
            else:
                raise ValueError(
                    f"Expected either a tuple with {len(inputs)} values "
                    f"or {len(inputs)} values across two arguments.")
        else:
            return inputs[0]
    return inputs


def resolve_tuple(sequence):
    """Resolves a sequence to a tuple."""
    if isinstance(sequence, np.ndarray):
        sequence = sequence.tolist()
    return tuple(i for i in sequence)


def has_nested_dicts(obj):
    """Returns whether a dictionary contains nested dicts."""
    return any(isinstance(i, dict) for i in obj.values())


def as_scalar(inp):
    """Converts an input value to a scalar."""
    if isinstance(inp, (int, float)):
        return inp
    if np.isscalar(inp):
        return inp.item()
    if isinstance(inp, np.ndarray):
        return inp.item()
    from agml.backend.tftorch import torch
    if isinstance(inp, torch.Tensor):
        return inp.item()
    from agml.backend.tftorch import tf
    if isinstance(inp, tf.Tensor):
        return inp.numpy()
    raise TypeError(f"Unsupported variable type {type(inp)}.")


def scalar_unpack(inp):
    """Unpacks a 1-d array into a list of scalars."""
    return [as_scalar(item) for item in inp]


def is_array_like(inp):
    """Determines if an input is a np.ndarray, torch.Tensor, or tf.Tensor."""
    if isinstance(inp, np.ndarray):
        return True
    from agml.backend.tftorch import torch
    if isinstance(inp, torch.Tensor):
        return True
    from agml.backend.tftorch import tf
    if isinstance(inp, tf.Tensor):
        return True
    return False


def shapes(seq):
    """Returns the shapes (or lengths) of all of the objects in the sequence."""
    try:
        return [getattr(obj, 'shape', len(obj)) for obj in seq]
    except:
        raise ValueError(f"One or more of the objects has no shape or length: {seq}.")


# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import json
import shutil
import difflib
import functools


@functools.lru_cache(maxsize = None)
def load_public_sources() -> dict:
    """Loads the public data sources JSON file."""
    with open(os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            '_assets/public_datasources.json')) as f:
        return json.load(f)


@functools.lru_cache(maxsize = None)
def load_citation_sources() -> dict:
    """Loads the citation sources JSON file."""
    with open(os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            '_assets/source_citations.json')) as f:
        return json.load(f)


@functools.lru_cache(maxsize = None)
def load_model_benchmarks() -> dict:
    """Loads the citation sources JSON file."""
    with open(os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            '_assets/model_benchmarks.json')) as f:
        return json.load(f)


def maybe_you_meant(name, msg, source = None) -> str:
    """Suggests potential correct spellings for an invalid name."""
    source = source if source is not None \
                    else load_public_sources().keys()
    suggestion = difflib.get_close_matches(name, source)
    if len(suggestion) == 0:
        return msg
    return msg + f" Maybe you meant: '{suggestion[0]}'?"


def copyright_print(name, location = None):
    """Prints out license/copyright info after a dataset download."""
    content = load_citation_sources()[name]
    license = content['license'] # noqa
    citation = content['citation']

    def _bold(msg): # noqa
        return '\033[1m' + msg + '\033[0m'

    if location is None:
        first_msg = "Citation information for " + _bold(name) + ".\n"
    else:
        first_msg = "You have just downloaded " + _bold(name) + ".\n"

    _LICENSE_TO_URL = {
        'CC BY-SA 4.0': 'https://creativecommons.org/licenses/by-sa/4.0/',
        'CC BY-SA 3.0': 'https://creativecommons.org/licenses/by-sa/3.0/',
        'CC BY-NC 3.0': 'https://creativecommons.org/licenses/by-nc/3.0/',
        'CC BY-NC SA 3.0': 'https://creativecommons.org/licenses/by-nc/3.0/',
        'MIT': 'https://opensource.org/licenses/MIT',
        'GPL-3.0': 'https://opensource.org/licenses/GPL-3.0',
        'US Public Domain': 'https://www.usa.gov/government-works'}
    if license == '':
        license_msg = "This dataset has " \
                      + _bold("no license") + ".\n"
    else:
        license_msg = "This dataset is licensed under the " \
                      + _bold(license) + " license.\n"
        license_msg += "To learn more, visit: " \
                       + _LICENSE_TO_URL[license] + "\n"

    if citation == '':
        citation_msg = "This dataset has no associated citation."
    else:
        citation_msg = "When using this dataset, please cite the following:\n\n"
        citation_msg += citation

    docs = load_public_sources()[name]['docs_url']
    docs_msg = ("\nYou can find additional information about "
                "this dataset at:\n" + docs)

    columns = shutil.get_terminal_size((80, 24)).columns
    max_print_length = max(min(
        columns, max([len(i) for i in [
            *citation_msg.split('\n'), *license_msg.split('\n')]])), columns)
    print('\n' + '=' * max_print_length)
    print(first_msg)
    print(license_msg)
    print(citation_msg)
    print(docs_msg)
    print("\nThis message will " + _bold("not") + " be automatically shown\n" 
          "again. To view this message again, in an AgMLDataLoader\n" +
          "run `loader.info.citation_summary()`. Otherwise, you\n" +
          "can use `agml.data.source(<name>).citation_summary().`")

    if location is not None:
        print(f"\nYou can find your dataset at {location}.")
    print('=' * max_print_length)
    sys.stdout.flush()








# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import glob
import shutil
import tempfile
import subprocess as sp
from datetime import datetime as dt

from agml.synthetic.config import HELIOS_PATH


# Helios build and compilation paths.
PROJECT_PATH = os.path.join(
    HELIOS_PATH, 'projects/SyntheticImageAnnotation')
HELIOS_BUILD = os.path.join(
    PROJECT_PATH, 'build')
HELIOS_EXECUTABLE = os.path.join(
    PROJECT_PATH, 'build', 'SyntheticImageAnnotation')

# Helios parameter paths.
XML_PATH = os.path.join(PROJECT_PATH, 'xml')


def _compile_helios_default():
    """Compiles the default Helios library upon installation and update."""
    if not os.path.exists(PROJECT_PATH):
        raise NotADirectoryError(
            f"The expected project path {PROJECT_PATH} does not exist. "
            f"There may have been an error in installing Helios. Try "
            f"re-installing it, or raise an issue with the AgML team.")

    # If there is an existing compiled version of Helios, then cache it
    # temporarily until the new compilation is complete. If an error is
    # encountered during the compilation, then we want to have a backup.
    os.makedirs(HELIOS_BUILD, exist_ok = True)
    temp_dir = tempfile.TemporaryDirectory()
    helios_temp_dir = os.path.join(temp_dir.name, 'helios_build')
    shutil.move(HELIOS_BUILD, helios_temp_dir)

    # Check that LiDAR compilation is disabled by default. We need to
    # store the original pulled contents of the CMake file, however, so
    # that we can revert it once it's been edited for the purposes of
    # compilation. This is so that new versions of Helios can be pulled
    # from Git without any clashes due to the CMake file having been edited.
    cmake_file = os.path.join(PROJECT_PATH, 'CMakeLists.txt')
    with open(cmake_file, 'r') as f:
        default_cmake_contents = f.read()
    cmake_contents = re.sub('lidar;', '', default_cmake_contents)
    with open(cmake_file, 'w') as f:
        f.write(cmake_contents)

    # Construct arguments for the compilation.
    cmake_args = ['cmake', '..', '-G', 'Unix Makefiles',
                  f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={HELIOS_BUILD}']
    make_args = ['cmake', '--build', '.']

    # Create the log file and clear the existing one.
    log_file = os.path.expanduser(
        f"~/.agml/.helios_compilation_log"
        f"-{dt.now().strftime('%Y%m%d-%H%M%S')}.log")
    try:
        existing_log = glob.glob(os.path.expanduser(
            '~/.agml/.helios_compilation_log-*.log'))[0]
        os.remove(existing_log)
    except:
        pass

    # Create the build directory again, since it has been moved.
    os.makedirs(HELIOS_BUILD, exist_ok = True)

    # Compile the CMake files.
    cmake_log = ""
    sys.stderr.write("Compiling Helios with CMake.\n\n")
    cmake_process = sp.Popen(cmake_args, stdout = sp.PIPE, cwd = HELIOS_BUILD,
                             stderr = sp.STDOUT, universal_newlines = True)
    for line in iter(cmake_process.stdout.readline, ""):
        cmake_log += line
        sys.stderr.write(line)
    cmake_process.stdout.close()
    with open(log_file, 'a') as f:
        f.write(cmake_log)
    cmake_return = cmake_process.wait()
    if cmake_return != 0:
        sys.stderr.write(
            f'\nEncountered an error when attempting to compile '
            f'Helios with CMake. Please report this traceback to '
            f'the AgML team. A full traceback of the compilation '
            f'process can be found at "{log_file}".')
        _compilation_failed(helios_temp_dir, temp_dir)
    sys.stdout.write('\n')
    sys.stderr.write('\n')

    # Generate the main executable.
    cmake_log = "\n"
    sys.stderr.write("Building Helios executable with CMake.\n\n")
    make_process = sp.Popen(make_args, stdout = sp.PIPE, cwd = HELIOS_BUILD,
                            stderr = sp.STDOUT, universal_newlines = True)
    for line in iter(make_process.stdout.readline, ""):
        cmake_log += line
        sys.stderr.write(line)
    make_process.stdout.close()
    with open(log_file, 'a') as f:
        f.write(cmake_log)
    make_return = make_process.wait()
    if make_return != 0:
        sys.stderr.write(
            f'\nEncountered an error when attempting to compile '
            f'Helios with CMake. Please report this traceback to '
            f'the AgML team. A full traceback of the compilation '
            f'process can be found at "{log_file}".')
        _compilation_failed(helios_temp_dir, temp_dir)

    # Print the final message.
    sys.stdout.write('\n')
    sys.stderr.write('\nHelios compilation successful!')


def _compile_helios_executable_only():
    """Compiles only the Helios executable."""
    if not os.path.exists(PROJECT_PATH):
        raise NotADirectoryError(
            f"The expected project path {PROJECT_PATH} does not exist. "
            f"There may have been an error in installing Helios. Try "
            f"re-installing it, or raise an issue with the AgML team.")

    # Construct arguments for the compilation.
    make_args = ['cmake', '--build', '.']

    # Create the log file and clear the existing one.
    log_file = os.path.expanduser(
        f"~/.agml/.helios_compilation_log"
        f"-{dt.now().strftime('%Y%m%d-%H%M%S')}.log")
    try:
        existing_log = glob.glob(os.path.expanduser(
            '~/.agml/.helios_compilation_log-*.log'))[0]
        os.remove(existing_log)
    except:
        pass

    # Generate the main executable.
    cmake_log = "\n"
    sys.stderr.write("Building Helios executable with CMake.\n\n")
    make_process = sp.Popen(make_args, stdout = sp.PIPE, cwd = HELIOS_BUILD,
                            stderr = sp.STDOUT, universal_newlines = True)
    for line in iter(make_process.stdout.readline, ""):
        cmake_log += line
        sys.stderr.write(line)
    make_process.stdout.close()
    with open(log_file, 'a') as f:
        f.write(cmake_log)
    make_return = make_process.wait()
    if make_return != 0:
        sys.stderr.write(
            f'\nEncountered an error when attempting to compile '
            f'Helios with CMake. Please report this traceback to '
            f'the AgML team. A full traceback of the compilation '
            f'process can be found at "{log_file}".')

    # Print the final message.
    sys.stdout.write('\n')
    sys.stderr.write('\nHelios compilation successful!')


def recompile_helios(executable_only = False):
    """Recompiles the Helios library with the set parameters.

    This method can be used by the user in order to recompile Helios, if, for
    instance, an error was encountered during prior compilation, or a local edit
    is made, without having to wait for the default recompilation every 48 hours.

    If just the `generate.cpp` file itself is edited, then you can pass the
    parameter `executable_only` as `True`, which will result in a much quicker
    compilation as just the generation file itself is edited.
    """
    if executable_only:
        _compile_helios_executable_only()
    else:
        _compile_helios_default()


def _compilation_failed(helios_temp_dir, temp_dir):
    """Cleanup after a failed Helios compilation."""
    shutil.rmtree(HELIOS_BUILD)
    shutil.move(helios_temp_dir, HELIOS_BUILD)
    temp_dir.cleanup()
    sys.exit(1)







# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


def _is_valid_file(file):
    """Returns whether a file is valid.

    This means that it is a file and not a directory, but also that
    it isn't an unnecessary dummy file like `.DS_Store` on MacOS.
    """
    if os.path.isfile(file):
        if not file.startswith('.git') and file not in ['.DS_Store']:
            return True
    return False


def get_file_list(fpath, ext = None, full_paths = True):
    """Gets a list of files from a path."""
    base_list = [os.path.join(fpath, f) if full_paths else f
                 for f in os.listdir(fpath)
                 if _is_valid_file(os.path.join(fpath, f))]
    if ext is not None:
        if isinstance(ext, str):
            return [f for f in base_list if f.endswith(ext)]
        else:
            return [f for f in base_list if
                    any([f.endswith(i) for i in ext])]
    return base_list


def get_dir_list(filepath):
    """Get a list of directories from a path."""
    return [f for f in os.listdir(filepath)
            if os.path.isdir(os.path.join(filepath, f))]


def nested_dir_list(fpath):
    """Returns a nested list of directories from a path."""
    dirs = []
    for f in os.scandir(fpath): # type: os.DirEntry
        if f.is_dir():
            if not os.path.basename(f.path).startswith('.'):
                dirs.append(os.path.join(fpath, f.path))
    if len(dirs) != 0:
        for dir_ in dirs:
            dirs.extend(nested_dir_list(dir_))
    return dirs


def nested_file_list(fpath):
    """Returns a nested list of files from a path."""
    files = []
    dirs = nested_dir_list(fpath)
    for dir_ in dirs:
        files.extend([os.path.join(dir_, i) for i in os.listdir(dir_)
                      if _is_valid_file(os.path.join(dir_, i))])
    return files


def create_dir(dir_):
    """Creates a directory (or does nothing if it exists)."""
    os.makedirs(dir_, exist_ok = True)


def recursive_dirname(dir_, level = 1):
    """Returns a recursive dirname for the number of levels provided."""
    if level == 0:
        return dir_
    return recursive_dirname(os.path.dirname(dir_), level - 1)


def is_image_file(file):
    """Returns whether a file is an image file."""
    if not isinstance(file, (str, bytes, os.PathLike)):
        return False
    end = os.path.splitext(file)[-1][1:]
    return end.lower() in ['jpg', 'jpeg', 'png', 'bmp', 'tiff']




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import re
import sys
import yaml
import collections
from typing import Iterable

from agml.framework import AgMLSerializable
import agml.utils.logging as logging
from agml.utils.general import has_nested_dicts
from agml.utils.data import (
    load_public_sources, load_citation_sources, maybe_you_meant, copyright_print
)


# Named tuples which are used by the metadata.
Tasks = collections.namedtuple('Tasks', ['ml', 'ag'])
Location = collections.namedtuple('Location', ['continent', 'country'])
ImageStats = collections.namedtuple('ImageStats', ['mean', 'std'])


def make_metadata(name, meta = None):
    """Creates the metadata object for the dataset.

    For datasets in the AgML public data repository, this simply returns
    a regular `DatasetMetadata` object with a full set of information for
    the dataset. Otherwise, it returns a `CustomDatasetMetadata` object,
    containing only the provided information in a `meta` dictionary.
    """
    if isinstance(name, (list, tuple, set, frozenset)): # a single dataset
        name = next(iter(name))
    if name in load_public_sources().keys():
        return DatasetMetadata(name)
    return CustomDatasetMetadata(name, meta)


class _MetadataDict(dict):
    """Dictionary subclass that throws a custom error for metadata accesses."""
    def __init__(self, *args, dataset = None, **kwargs):
        if dataset is None:
            raise ValueError(
                "Cannot instantiate metadata dictionary without the dataset name.")
        super(_MetadataDict, self).__init__(*args, **kwargs)
        self._dataset = dataset
        self._custom = kwargs.get('custom', False)

    def __getitem__(self, item):
        try:
            return super(_MetadataDict, self).__getitem__(item)
        except KeyError:
            if self._custom:
                raise KeyError(
                    f"The provided custom dataset '{self._dataset}' has no "
                    f"associated metadata '{item}'. Please provide this when"
                    f"instantiating the loader if you want to use it.")
            raise KeyError(
                f"The dataset '{self._dataset}' is missing metadata '{item}'. "
                f"Please bring this issue to the attention of the AgML team.")


class DatasetMetadata(AgMLSerializable):
    """Stores metadata about a certain AgML dataset.

    When loading in a dataset using the `AgMLDataLoader`, the "info"
    parameter of the class will contain a `DatasetMetadata` object,
    which will expose metadata including (but not limited to):

    - The original dataset source,
    - The location that dataset images were captured,
    - The image and sensor modality, and
    - The data and annotation formats.

    Generally, this can be used as follows:

    > ds = agml.AgMLDataLoader('<dataset-name>')
    > ds.info
    <dataset-name>
    > ds.info.annotation_format
    <annotation-format>

    Most attributes of the dataset will be directly available as a
    property of this class, but any additional info that is not can
    be accessed by treating the `info` object as a dictionary.
    """
    serializable = frozenset(('name', 'metadata', 'citation_meta'))
    is_custom_dataset = False

    def __init__(self, name):
        self._load_source_info(name)

    def __repr__(self):
        return self._name

    def __str__(self):
        return self._name

    def __fspath__(self):
        return self._name

    def __eq__(self, other):
        if isinstance(other, DatasetMetadata):
            return self._name == other._name
        if isinstance(other, str):
            return self._name == other
        return False

    @property
    def data(self):
        return self._metadata

    def _load_source_info(self, name):
        """Loads the data source metadata into the class."""
        if isinstance(name, DatasetMetadata):
            name = name.name

        source_info = load_public_sources()
        if name not in source_info.keys():
            if name.replace('-', '_') not in source_info.keys():
                msg = f"Received invalid public source: '{name}'."
                msg = maybe_you_meant(name, msg)
                raise ValueError(msg)
            else:
                logging.log(
                    f"Interpreted dataset '{name}' as '{name.replace('-', '_')}.'")
                name = name.replace('-', '_')
        self._name = name
        self._metadata = _MetadataDict(
            **source_info[name], dataset = name)

        # Load citation information.
        self._citation_meta = _MetadataDict(
            **load_citation_sources()[name], dataset = name)

    def __getattr__(self, key):
        try:
            # Some weird behavior with lookups can happen.
            return object.__getattribute__(self, key)
        except:
            if key in self._metadata.keys():
                return self._metadata[key]
            raise AttributeError(
                maybe_you_meant(
                    key, f"Received invalid info parameter: '{key}'.",
                    source = self._metadata.keys()))

    def __getitem__(self, key):
        return getattr(self, key)

    @property
    def name(self):
        return self._name

    @property
    def num_images(self):
        return int(float(self._metadata['n_images']))

    @property
    def tasks(self):
        """Returns the ML and Agriculture tasks for this dataset."""
        ml_task, ag_task = self._metadata['ml_task'], self._metadata['ag_task']
        return Tasks(ml = ml_task, ag = ag_task)

    @property
    def location(self):
        """Returns the continent and country in which the dataset was made."""
        continent, country = self._metadata['location'].values()
        return Location(continent = continent, country = country)

    @property
    def image_stats(self):
        """Returns the mean and standard deviation of the RGB images."""
        mean, std = self._metadata['stats'].values()
        return ImageStats(mean = mean, std = std)

    @property
    def sensor_modality(self):
        return self._metadata['sensor_modality']

    @property
    def image_format(self):
        return self._metadata['input_data_format']

    @property
    def annotation_format(self):
        return self._metadata['annotation_format']

    @property
    def docs(self):
        return self._metadata['docs_url']

    @property
    def num_to_class(self):
        mapping = self._metadata['classes']
        if has_nested_dicts(mapping):
            out = {}
            for class_type in mapping.keys():
                if isinstance(mapping[class_type], dict):
                    nums = [int(float(i)) for i in mapping[class_type.keys()]]
                    out[class_type] = dict(zip(nums, mapping[class_type].values()))
                else:
                    out[class_type] = mapping[class_type]
            return out
        nums = [int(float(i)) for i in mapping.keys()]
        return dict(zip(nums, mapping.values()))

    @property
    def class_to_num(self):
        mapping = self._metadata['classes']
        if has_nested_dicts(mapping):
            out = {}
            for class_type in mapping.keys():
                if isinstance(mapping[class_type], dict):
                    nums = [int(float(i)) for i in mapping[class_type].keys()]
                    out[class_type] = dict(zip(mapping[class_type].values(), nums))
                else:
                    out[class_type] = mapping[class_type]
            return out
        nums = [int(float(i)) for i in mapping.keys()]
        return dict(zip(mapping.values(), nums))

    @property
    def classes(self):
        classes = self._metadata['classes']
        if has_nested_dicts(classes):
            return {k: list(d.values()) for k, d in classes.items()}
        return list(classes.values())

    @property
    def num_classes(self):
        return len(self.classes)

    @property
    def license(self):
        if self._citation_meta['license'] == '':
            return None
        return self._citation_meta['license']

    @property
    def citation(self):
        if self._citation_meta['citation'] == '':
            return None
        return self._citation_meta['citation']

    @property
    def external_image_sources(self):
        return self._metadata['external_image_sources']

    def summary(self):
        """Prints out a summary of the dataset information.

        For all of the properties of information defined in this
        metadata class, this method will print out a summary of all
        of the information in a visually understandable table. Note
        that this does not return anything, so it shouldn't be called
        as `print(loader.info.summary())`, just `loader.info.summary()`.
        """
        def _bold(msg):
            return '\033[1m' + msg + '\033[0m'
        def _bold_yaml(msg): # noqa
            return '<|>' + msg + '<|>'

        _SWITCH_NAMES = {
            'ml_task': "Machine Learning Task",
            'ag_task': "Agricultural Task",
            'real_synthetic': 'Real Or Synthetic',
            'n_images': "Number of Images",
            'docs_url': "Documentation"
        }

        formatted_metadata = {}
        for key, value in self._metadata.items():
            name = key.replace('_', ' ').title()
            if key in _SWITCH_NAMES.keys():
                name = _SWITCH_NAMES[key]
            if name == 'Crop Types':
                value = {int(k): v for k, v in value.items()}
            if name == 'Number of Images':
                value = int(value)
            formatted_metadata[_bold_yaml(name)] = value

        stream = io.StringIO()
        yaml.dump(formatted_metadata, stream, sort_keys = False)
        content = stream.getvalue()
        content = re.sub('<\\|>(.*?)<\\|>', _bold(r'\1'), content)
        header = '=' * 20 + ' DATASET SUMMARY ' + '=' * 20
        print(header)
        print(_bold("Name") + f": {self._name}")
        print(content, end = '')
        print('=' * 57)
        sys.stdout.flush()

    def citation_summary(self):
        """Prints out a summary of the citation information of the dataset.

        This message is the same as is displayed when the dataset is
        initially downloaded, and contains information about the dataset
        license and associated citation (if either exist).
        """
        copyright_print(self._name)


class CustomDatasetMetadata(DatasetMetadata):
    """A metadata wrapper for information given for custom datasets.

    When using a custom dataset in the `AgMLDataLoader`, as there is no
    metadata in the AgML public data source information file, custom
    metadata needs to be passed to the loader instantiation. However,
    this metadata does not necessarily need to be complete to the full
    extent of the rest of the metadata, and so this class is used
    to wrap the provided metadata and infer other arguments instead.
    """
    serializable = frozenset(("name", "metadata"))
    is_custom_dataset = True

    def __init__(self, name, meta): # noqa
        if meta is None:
            msg = f"Received invalid public source: '{name}'."
            out_msg = maybe_you_meant(name, msg)
            if out_msg == msg:
                raise ValueError("Expected metadata when creating a custom loader, got None.")
            else:
                raise ValueError(out_msg)
        self._load_info(name, meta)

    def _load_info(self, name, meta):
        """Loads the provided metadata into the class."""
        # Check that the necessary categories (the task and a list of classes)
        # have been provided, otherwise the loader will not be able to function.
        if 'task' not in meta:
            raise ValueError("Expected a `task` when instantiating a custom loader.")
        if 'classes' not in meta:
            raise ValueError("Expected a list of classes when instantiating a custom loader.")

        # Create a custom dictionary of metadata.
        self._name = name
        self._metadata = _MetadataDict(
            {'ml_task': meta['task'], 'ag_task': meta.get('ag_task', None),
             **meta}, dataset = name)
        self._metadata['classes'] = {str(i + 1): c for i, c in enumerate(meta['classes'])}

        # There is no citation information necessary for custom datasets.
        self._citation_meta = None

    @property
    def license(self):
        raise ValueError("There is no citation information for custom datasets.")

    @property
    def citation(self):
        raise ValueError("There is no citation information for custom datasets.")

    def citation_summary(self): # no citations for custom datasets.
        raise ValueError("There is no citation information for custom datasets.")








# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from enum import Enum
from numbers import Number
from dataclasses import dataclass, fields, asdict
from typing import List, Union, Sequence

from agml.framework import AgMLSerializable
from agml.synthetic.config import load_default_helios_configuration
from agml.synthetic.tools import generate_camera_positions


class AnnotationType(Enum):
    """The type of annotation to use in generation."""
    object_detection: str = "object_detection"
    semantic_segmentation: str = "semantic_segmentation"
    instance_segmentation: str = "instance_segmentation"
    none: str = "none"


class SimulationType(Enum):
    """The simulation render (RGB vs. LiDAR) that is generated."""
    RGB: str = "rgb"
    LiDAR: str = "lidar"


@dataclass(repr = False)
class Parameters:
    """Base class for parameters, to enable runtime type checks."""
    def __post_init__(self):
        # Remove all parameters which don't belong to this class. We know
        # they don't belong if they are still `None` after initialization.
        self_dict = self.__dict__.copy()
        for attr in self_dict:
            if self.__dict__[attr] is None:
                delattr(self, attr)
        self._block_new_attributes = True

    def __repr__(self):
        # This is custom-defined to exclude optional unused attributes.
        defined_values = (
            (f.name, getattr(self, f.name))
            for f in fields(self)
            if getattr(self, f.name) != f.default)
        value_repr = ", ".join(f"{name}={value}" for name, value in defined_values)
        return f"{self.__class__.__qualname__}({value_repr})"

    def __setattr__(self, key, value):
        # Don't allow the assignment of new attributes.
        if key not in self.__dict__.keys():
            if not hasattr(self, '_block_new_attributes'):
                super().__setattr__(key, value)
                return
            raise AttributeError(f"Cannot assign new attributes '{key}' "
                                 f"to class {self.__class__.__name__}.")

        # Check if the type of the value matches that of the key.
        try:
            annotation = self.__annotations__[key].__origin__
        except AttributeError:
            # enables type checks for subscripted generics
            annotation = (int, float)
        if not isinstance(value, annotation):
            raise TypeError(
                f"Expected a value of type ({annotation}) for attribute "
                f"'{key}', instead got '{value}' of type ({type(value)}).")
        super().__setattr__(key, value)


@dataclass(repr = False)
class CanopyParameters(Parameters):
    """Stores canopy-specific parameters for Helios.

    This class is a super-parameter class which contains all of the parameters
    across all of the canopies within Helios, though each of the individual canopies
    may only use a fraction of these specific canopies. This is to ensure that
    different canopy parameters auto-complete when actually writing code.
    """
    leaf_length: Number                   = None
    leaf_width: Number                    = None
    leaf_size: Number                     = None
    leaf_subdivisions: List[Number]       = None
    leaf_texture_file: str                = None
    leaf_color: str                       = None
    leaf_angle_distribution: str          = None
    leaf_area_index: Number               = None
    leaf_area_density: Number             = None
    leaf_spacing_fraction: Number         = None
    leaflet_length: Number                = None
    stem_color: List[Number]              = None
    stem_subdivisions: Number             = None
    stems_per_plant: Number               = None
    stem_radius: Number                   = None
    stem_length: Number                   = None
    plant_height: Number                  = None
    grape_radius: Number                  = None
    grape_color: List[Number]             = None
    grape_subdivisions: Number            = None
    fruit_color: List[Number]             = None
    fruit_radius: Number                  = None
    fruit_subdivisions: Number            = None
    fruit_texture_file: os.PathLike       = None
    wood_texture_file: os.PathLike        = None
    wood_subdivisions: Number             = None
    clusters_per_stem: Number             = None
    plant_spacing: Number                 = None
    row_spacing: Number                   = None
    level_spacing: Number                 = None
    plant_count: List[Number]             = None
    canopy_origin: List[Number]           = None
    canopy_rotation: Number               = None
    canopy_height: Number                 = None
    canopy_extent: List[Number]           = None
    canopy_configuration: str             = None
    base_height: Number                   = None
    crown_radius: Number                  = None
    cluster_radius: Number                = None
    cluster_height_max: Number            = None
    trunk_height: Number                  = None
    trunk_radius: Number                  = None
    cordon_height: Number                 = None
    cordon_radius: Number                 = None
    cordon_spacing: Number                = None
    shoot_length: Number                  = None
    shoot_radius: Number                  = None
    shoots_per_cordon: Number             = None
    shoot_angle: Number                   = None
    shoot_angle_tip: Number               = None
    shoot_angle_base: Number              = None
    shoot_color: List[Number]             = None
    shoot_subdivisions: List[Number]      = None
    pod_color: List[Number]               = None
    pod_subdivisions: List[Number]        = None
    pod_length: None                      = None
    germination_probability: Number       = None
    needle_width: Number                  = None
    needle_length: Number                 = None
    needle_color: List[Number]            = None
    needle_subdivisions: List[Number]     = None
    branch_length: Number                 = None
    branches_per_level: Number            = None
    buffer: str                           = None

    # sorghum options
    sorghum_stage: Number                 = None
    s1_stem_length: Number                = None
    s1_stem_radius: Number                = None
    s1_stem_subdivisions: Number          = None
    s1_leaf_size1: List[Number]           = None
    s1_leaf_size2: List[Number]           = None
    s1_leaf_size3: List[Number]           = None
    s1_leaf1_angle: Number                = None
    s1_leaf2_angle: Number                = None
    s1_leaf3_angle: Number                = None
    s1_leaf_subdivisions: List[Number]    = None
    s1_leaf_texture_file: str             = None
    s2_stem_length: Number                = None
    s2_stem_radius: Number                = None
    s2_stem_subdivisions: Number          = None
    s2_leaf_size1: List[Number]           = None
    s2_leaf_size2: List[Number]           = None
    s2_leaf_size3: List[Number]           = None
    s2_leaf_size4: List[Number]           = None
    s2_leaf_size5: List[Number]           = None
    s2_leaf1_angle: Number                = None
    s2_leaf2_angle: Number                = None
    s2_leaf3_angle: Number                = None
    s2_leaf4_angle: Number                = None
    s2_leaf5_angle: Number                = None
    s2_leaf_subdivisions: List[Number]    = None
    s2_leaf_texture_file: str             = None
    s3_stem_length: Number                = None
    s3_stem_radius: Number                = None
    s3_stem_subdivisions: Number          = None
    s3_leaf_size: List[Number]            = None
    s3_leaf_subdivisions: List[Number]    = None
    s3_number_of_leaves: Number           = None
    s3_mean_leaf_angle: Number            = None
    s3_leaf_texture_file: str             = None
    s4_stem_length: Number                = None
    s4_stem_radius: Number                = None
    s4_stem_subdivisions: Number          = None
    s4_panicle_size: List[Number]         = None
    s4_panicle_subdivisions: Number       = None
    s4_seed_texture_file: str             = None
    s4_leaf_size: List[Number]            = None
    s4_leaf_subdivisions: List[Number]    = None
    s4_number_of_leaves: Number           = None
    s4_mean_leaf_angle: Number            = None
    s4_leaf_texture_file: str             = None
    s5_stem_length: Number                = None
    s5_stem_radius: Number                = None
    s5_stem_bend: Number                  = None
    s5_stem_subdivisions: Number          = None
    s5_panicle_size: List[Number]         = None
    s5_panicle_subdivisions: Number       = None
    s5_seed_texture_file: str             = None
    s5_leaf_size: List[Number]            = None
    s5_leaf_subdivisions: List[Number]    = None
    s5_number_of_leaves: Number           = None
    s5_mean_leaf_angle: Number            = None
    s5_leaf_texture_file: str             = None


@dataclass(repr = False)
class CameraParameters(Parameters):
    """Stores camera parameters for Helios."""
    image_resolution: List[Number]  = None
    camera_position: List[Number]   = None
    camera_lookat: List[Number]     = None

    def generate_positions(self,
                           camera_type: str,
                           num_views: int,
                           origin: List[Union[int, float]] = None,
                           camera_spacing: int = 2,
                           crop_distance: int = 4,
                           height: int = 1):
        """Generates camera positions from the input environment parameters."""
        self.camera_position, self.camera_lookat = generate_camera_positions(
            camera_type = camera_type, num_views = num_views,
            origin = origin, camera_spacing = camera_spacing,
            crop_distance = crop_distance, height = height)


@dataclass(repr = False)
class LiDARParameters(Parameters):
    """Stores LiDAR parameters for Helios."""
    origin: List[Number]    = None
    size: List[Number]      = None
    thetaMin: Number        = None
    thetaMax: Number        = None
    phiMin: Number          = None
    phiMax: Number          = None
    exitDiameter: Number    = None
    beamDivergence: Number  = None
    ASCII_format: str       = None


class HeliosOptions(AgMLSerializable):
    """Stores a set of parameter options for a `HeliosDataGenerator`.

    The primary exposed options are the `canopy_parameters` (as well as similar
    options for camera position and LiDAR, `camera/lidar_parameters' specifically),
    and the actual data generation parameters, such as the `annotation_format`.

    The `HeliosOptions` is instantiated with the name of the canopy type that you
    want to generate; from there, the parameters and ranges can be accessed through
    properties, which are set to their default values as loaded from the Helios
    configuration upon instantiating the class.

    This `HeliosOptions` object, once configured, can then be passed directly to
    a `HeliosDataGenerator` upon instantiation, and be used to generate synthetic
    data according to its specification. The options can be edited and then the
    generation process run again to obtain a new set of data.

    Parameters
    ----------
    canopy : str
        The type of plant canopy to be used in generation.
    """
    serializable = frozenset(('canopy', 'canopy_parameters',
                              'camera_parameters', 'lidar_parameters',
                              'annotation_type', 'simulation_type', 'labels'))
    
    def __new__(cls, *args, **kwargs):
        # Run the configuration check. Notice that we run this here because
        # users may not necessarily want to use Helios on first installing
        # AgML, but the `synthetic` module is loaded into the main AgML
        # module as in the top-level `__init__.py` file. By putting the check
        # here, it ensures that Helios is only installed if and when data is
        # actually being generated, and not just by a standard import.
        from .config import _check_helios_installation
        _check_helios_installation()
        del _check_helios_installation

        # The default configuration parameters are loaded directly from
        # the `helios_config.json` file which is constructed each time
        # Helios is installed or updated.
        cls._default_config = load_default_helios_configuration()
        return super(HeliosOptions, cls).__new__(cls)

    def __init__(self, canopy = None):
        # Check that the provided canopy is valid.
        self._initialize_canopy(canopy)

        # Initialize the default data generation parameters.
        self._annotation_type = AnnotationType.object_detection
        self._simulation_type = SimulationType.RGB
        self._labels = ['leaves']

    def _initialize_canopy(self, canopy):
        """Initializes Helios options from the provided canopy."""
        if canopy not in self._default_config['canopy']['types']:
            raise ValueError(
                f"Received invalid canopy type '{canopy}', expected "
                f"one of: {self._default_config['canopy']['types']}.")
        self._canopy = canopy

        # Get the parameters and ranges corresponding to the canopy type.
        self._canopy_parameters = \
            CanopyParameters(**self._default_config['canopy']['parameters'][canopy])
        self._camera_parameters = \
            CameraParameters(**self._default_config['camera']['parameters'])
        self._lidar_parameters = \
            LiDARParameters(**self._default_config['lidar']['parameters'])

    @property
    def canopy(self) -> CanopyParameters:
        return self._canopy_parameters

    @property
    def camera(self) -> CameraParameters:
        return self._camera_parameters

    @property
    def lidar(self) -> LiDARParameters:
        return self._lidar_parameters
    
    @property
    def annotation_type(self) -> str:
        return self._annotation_type
    
    @annotation_type.setter
    def annotation_type(self, value: Union[AnnotationType, str]):
        self._annotation_type = AnnotationType(value)

    @property
    def simulation_type(self) -> str:
        return self._simulation_type

    @simulation_type.setter
    def simulation_type(self, value: Union[SimulationType, str]):
        self._simulation_type = SimulationType(value)

    @property
    def labels(self) -> list:
        return self._labels

    @labels.setter
    def labels(self, value: Sequence):
        if len(value) == 0:
            raise ValueError("You cannot have no labels, choose a combination "
                             "of `trunks`, `fruits`, `branches`, and `leaves`.")
        self._labels = value

    def reset(self):
        """Resets the parameters to the default for the canopy."""
        self._initialize_canopy(self._canopy)

    @staticmethod
    def _asdict(obj):
        return {k: v for k, v in asdict(obj).items() if v is not None}

    def _to_dict(self):
        """Returns the options in dictionary format."""
        return {
            'canopy': self._asdict(self._canopy_parameters),
            'camera': self._asdict(self._camera_parameters),
            'lidar': self._asdict(self._lidar_parameters)
        }





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

import numpy as np

from agml.framework import AgMLSerializable
from agml.utils.image import needs_batch_dim
from agml.backend.tftorch import (
    tf, torch, set_backend, get_backend,
    user_changed_backend, StrictBackendError,
    _convert_image_to_torch, is_array_like
)


class TrainState(Enum):
    NONE = None
    TF = 'tf'
    TORCH = 'torch'
    EVAL = 'eval'
    EVAL_TF = 'eval-tf'
    EVAL_TORCH = 'eval-torch'
    FALSE = False


# Shorthand form of the enum for testing explicit values.
t_ = TrainState


class TrainingManager(AgMLSerializable):
    """Controls the training state for the `AgMLDataLoader`.

    This manager is responsible for managing the backend system and
    training parameters of the `AgMLDataLoader`. In particular, it
    manages the data in different circumstances, such as when the
    data is in "train" and "eval" modes, or when it is set to be in
    a TensorFlow/PyTorch-compatible mode.

    This allows for more efficient compatibility management with the
    different backends, for instance dynamic tensor conversion and
    image formatting to account for channels_first vs. channels_last,
    as well as being able to manage whether certain preprocessing steps
    should be applied, allowing for independent train and eval modes.
    """
    serializable = frozenset((
        'transform_manager', 'resize_manager',
        'state', 'task', 'remap_hook', 'name'))

    def __init__(self, transform_manager, resize_manager, task = None):
        # Update the general parameters for the loader.
        self._task = task
        self._name = resize_manager._dataset_name

        # The `TrainingManager` is responsible for applying the
        # actual transforms, and thus controls the `TransformManager`
        # and the `ImageResizeManager` and their relevant states.
        self._transform_manager = transform_manager
        self._resize_manager = resize_manager

        # The `state` the the loader is in is variable, and can be
        # changed during program execution (e.g., between 'train'
        # and 'eval', or even into a specific backend).
        #
        # The `state` determines the actual preprocessing functions
        # that it should apply, as it controls the managers.
        # This variable tracks that state and uses it to determine
        # which steps that it should apply.
        #
        # See the `update_state()` method to see the valid states.
        self._state: "TrainState" = TrainState.NONE

        # A hook for multi-dataset loaders.
        self._remap_hook = False

    @property
    def state(self):
        """Exposes the internal state."""
        return self._state

    def update_state(self, state):
        """Updates the state of the training manager.

        This updates the state which is used to determine what preprocessing
        steps, if any, are applied to the data. Valid states include:

        1. `None`: This means that all of the transforms and image resizing
           will be applied to the data, but no automatic batching or tensor
           conversion. This is also the default state if none is specified.
        2. 'tf': This is one of the two training modes. This applies transforms,
           image resizing, as well as automatic batching and tensor conversion.
        3. 'torch': This is one of the two training modes. This applies
           transforms, image resizing, tensor conversion, as well as image
           formatting, but no automatic batching.
        4. 'eval': By default, this only enables image resizing. However, if
           the loader is set to a 'tf' or a 'torch' state, and from here it
           is converted to 'eval', then it also keeps potential tensor conversion,
           automatic batching, and image formatting.
        5. `False`: This disables all preprocessing and simply returns the raw
           loaded images and annotations.

        There are certain caveats here. Specifically, if you want to use 'eval'
        mode but maintain the tensor conversion and other related features, you
        need to use the following order of methods.

        > loader.as_keras_sequence() # or loader.as_torch_dataset()
        > loader.eval()

        If you want to completely disable preprocessing but keep the image
        resizing, then you need to first disable all preprocessing, then
        set the loader in 'eval' mode.

        > loader.as_keras_sequence() # or loader.as_torch_dataset()
        > loader.disable_preprocessing()
        > loader.eval()

        To re-enable preprocessing at this point, e.g., just the transforms
        and resizing but no tensor conversion or automatic batching, then use
        the following, which resets the train state of the loader.

        > loader.reset_processing()

        This enables the loader to track multiple states.
        """
        # Fully disable all preprocessing.
        if t_(state) == TrainState.FALSE:
            self._state = TrainState.FALSE

        # Set the correct 'eval' mode, based on the prior state.
        elif t_(state) == TrainState.EVAL:
            if self._state == TrainState.TF:
                self._state = TrainState.EVAL_TF
            elif self._state == TrainState.TORCH:
                self._state = TrainState.EVAL_TORCH
            else:
                self._state = TrainState.EVAL

        # Apply a 'tf' or 'torch' backend conversion.
        elif t_(state) == TrainState.TORCH:
            self._state = TrainState.TORCH
            if get_backend() == 'tf':
                if user_changed_backend():
                    raise StrictBackendError(
                        change = 'torch', obj = t_(state))
                set_backend('torch')
            self._resize_manager.assign('train-auto')
        elif t_(state) == TrainState.TF:
            self._state = TrainState.TF
            if get_backend() == 'torch':
                if user_changed_backend():
                    raise StrictBackendError(
                        change = 'tf', obj = t_(state))
                set_backend('tf')
            self._resize_manager.assign('train-auto')

        # Set the default conversion (`None`).
        elif t_(state) == TrainState.NONE:
            self._state = TrainState.NONE

    def _set_annotation_remap_hook(self, hook):
        """Used to modify class annotations for multi-dataset loaders."""
        self._remap_hook = hook

    def apply(self, obj, batch_state):
        """Applies preprocessing and conversions to the data contents.

        This method is responsible for actually loading and processing
        the data according to the training state, including loading the
        data, applying transforms and resizing, as well as the training
        management as described in the class. This is called by an
        enclosing `DataManager`.

        See the `TransformManager` and the `ImageResizeManager` for more
        information on the specific preprocessing applied there, and the
        `update_state()` method for more information on the training.
        """
        # Extract the raw contents from the `DataObject`.
        contents = obj.get()

        # If there is a hook to apply (for multi-dataset loaders),
        # then apply the hook before doing anything else.
        if self._remap_hook:
            contents = self._remap_hook(contents, self._name) # noqa

        # If the state is set to `False`, then just return the raw contents.
        if self._state is TrainState.FALSE:
            return contents

        # In any other case other than `False`, we resize the images.
        contents = self._resize_manager.apply(contents)

        # If we are in a training state or `None`, (so not an evaluation
        # state or `False`), then we apply the transforms to the images.
        if self._state not in [TrainState.EVAL,
                               TrainState.EVAL_TF,
                               TrainState.EVAL_TORCH]:
            contents = self._transform_manager.apply(contents)

        # If the images are not in a batch, then we convert them to tensors
        # here, otherwise, they will be converted when the batch is created.
        if not batch_state:
            contents = self._train_state_apply(contents)

        # Return the processed contents.
        return contents

    def _train_state_apply(self, contents):
        """Preprocesses the data according to the class's training state."""
        if self._state is TrainState.NONE:
            return contents
        elif self._state in [TrainState.TF, TrainState.EVAL_TF]:
            return self._tf_tensor_convert(contents, self._task)
        elif self._state in [TrainState.TORCH, TrainState.EVAL_TORCH]:
            return self._torch_tensor_convert(contents, self._task)
        return contents

    def make_batch(self, images, annotations):
        """Creates a batch of data out of processed images and annotations."""
        if self._state in [TrainState.NONE, TrainState.FALSE, TrainState.EVAL]:
            return images, annotations
        elif self._state in [TrainState.TF, TrainState.EVAL_TF]:
            return self._tf_tensor_batch_convert(
                (images, annotations), self._task)
        elif self._state in [TrainState.TORCH, TrainState.EVAL_TORCH]:
            return self._torch_tensor_batch_convert(
                (images, annotations), self._task)
        return images, annotations

    @staticmethod
    def _tf_tensor_image_convert(image):
        """Converts potential multi-image input dicts to tensors."""
        if isinstance(image, np.ndarray):
            return tf.constant(image)
        return {k: tf.constant(i) for k, i in image.items()}

    @staticmethod
    def _tf_tensor_image_batch_convert(batch):
        """Converts potential multi-image input batch dicts to tensor dicts."""
        if isinstance(batch, np.ndarray):
            return tf.stack(batch)
        return {k: tf.stack(b) for k, b in batch.items()}

    @staticmethod
    def _tf_tensor_convert(contents, task):
        """Converts contents to `tf.Tensor`s where possible."""
        # Convert the image and annotation to `tf.Tensor`s.
        image, annotation = contents
        image = TrainingManager._tf_tensor_image_convert(image)
        if task in ['image_classification',
                    'image_regression',
                    'semantic_segmentation']:
            if isinstance(annotation, (int, np.ndarray)):
                annotation = tf.constant(annotation)
            else:
                for k, v in annotation.items():
                    annotation[k] = tf.constant(v)
        elif task == 'object_detection':
            annotation = TrainingManager._tf_tensor_coco_convert(
                annotation)

        # Add a first-dimension batch to the image.
        if isinstance(image, tf.Tensor):
            if needs_batch_dim(image):
                image = tf.expand_dims(image, axis = 0)
        else:
            for k, v in image.items():
                if needs_batch_dim(v):
                    image[k] = tf.expand_dims(v, axis = 0)
        return image, annotation

    @staticmethod
    def _tf_tensor_batch_convert(contents, task):
        """Converts batch contents to `tf.Tensor`s where possible."""
        # This stacks the images and annotations together.
        images, annotations = contents
        images = TrainingManager._tf_tensor_image_batch_convert(images)
        if task == 'image_classification':
            annotations = tf.constant(annotations)
        elif task == 'image_regression':
            if isinstance(annotations, np.ndarray):
                annotations = tf.constant(annotations)
            else:
                for k, v in annotations.items():
                    annotations[k] = tf.constant(v)
        elif task == 'semantic_segmentation':
            annotations = tf.stack(annotations, axis = 0)
        elif task == 'object_detection':
            annotations = [TrainingManager._tf_tensor_coco_convert(
                a_set) for a_set in annotations]
        return images, annotations

    @staticmethod
    def _tf_tensor_coco_convert(contents):
        """Converts a COCO JSON dictionary to a `tf.Tensor`."""
        coco_tensor = {}
        for key, value in contents.items():
            coco_tensor[key] = tf.constant(value)
        return coco_tensor

    @staticmethod
    def _torch_tensor_image_convert(image):
        """Converts potential multi-image input dicts to tensors."""
        if is_array_like(image):
            return _convert_image_to_torch(image)
        return {k: _convert_image_to_torch(i) for k, i in image.items()}

    @staticmethod
    def _torch_tensor_image_batch_convert(batch):
        """Converts potential multi-image input batch dicts to tensor dicts."""
        if is_array_like(batch):
            return torch.stack([
                _convert_image_to_torch(image) for image in batch])
        return {k: TrainingManager.
            _torch_tensor_image_batch_convert(b) for k, b in batch.items()}

    @staticmethod
    def _torch_tensor_convert(contents, task):
        """Converts contents to `torch.Tensor`s where possible."""
        image, annotation = contents
        image = TrainingManager._torch_tensor_image_convert(image)
        if task in ['image_classification',
                    'image_regression']:
            if isinstance(annotation, (int, np.ndarray)):
                annotation = torch.tensor(annotation)
            else:
                for k, v in annotation.items():
                    annotation[k] = torch.tensor(v)
        elif task == 'semantic_segmentation':
            annotation = _convert_image_to_torch(annotation)
        elif task == 'object_detection':
            annotation = TrainingManager._torch_tensor_coco_convert(
                annotation)
        return image, annotation

    @staticmethod
    def _torch_tensor_batch_convert(contents, task):
        """Converts batch contents to `torch.Tensor`s where possible."""
        images, annotations = contents
        images = TrainingManager._torch_tensor_image_batch_convert(images)
        if task in ['image_classification',
                    'image_regression']:
            if isinstance(annotations, (int, np.ndarray)):
                annotations = torch.tensor(annotations)
            else:
                for k, v in annotations.items():
                    annotations[k] = torch.tensor(v)
        elif task == 'semantic_segmentation':
            annotations = torch.stack([
                _convert_image_to_torch(a) for a in annotations])
        elif task == 'object_detection':
            annotations = [TrainingManager._torch_tensor_coco_convert(
                a_set) for a_set in annotations]
        return images, annotations

    @staticmethod
    def _torch_tensor_coco_convert(contents):
        """Converts a COCO JSON dictionary to a `torch.Tensor`."""
        coco_tensor = {}
        for key, value in contents.items():
            if key == 'segmentation':
                value = np.empty(0)
            if not isinstance(value, torch.Tensor):
                coco_tensor[key] = torch.tensor(value)
            else:
                coco_tensor[key] = value
        return coco_tensor






